"""This module provides a simplified interface to the kubernetes module.

Attributes:
    kubectl (SysCmdRunner.run): A simple interface to the kubectl command line tool.
"""

# Import standard modules
from datetime import datetime as dt, timedelta
from pathlib import Path
from string import Template
from time import sleep
from typing import cast, Any, Callable, List, Optional, Type, TypeVar

# Import third-party modules
from kubernetes import config as k8s_config
from kubernetes.client import AppsV1Api, BatchV1Api, CoreV1Api, V1Namespace, V1ObjectMeta
from kubernetes.stream import stream as k8s_process
from yaml import safe_load as yaml_load

# Import internal modules
from .lang import DEFAULT_ENCODING, BatCaveError, BatCaveException, CommandResult, PathName
from .sysutil import SysCmdRunner

K8sObject = TypeVar('K8sObject', 'Namespace',  # pylint: disable=invalid-name
                    'Binding', 'Configmap', 'Endpoints', 'Event', 'Limitrange', 'Persistentvolumeclaim',
                    'Pod', 'Replicationcontroller', 'Resourcequota', 'Secret', 'Service', 'Serviceaccount',
                    'Daemonset', 'Deployment', 'Replicaset', 'Statefulset',
                    'Cronjob', 'Job')
XFORM_CLASS_NAMES = {'configmap': 'config_map', 'cronjob': 'cron_job',
                     'daemonset': 'daemon_set',
                     'limitrange': 'imit_range',
                     'persistentvolumeclaim': 'persistent_volume_claim',
                     'replicaset': 'replica_set',
                     'replicationcontroller': 'replication_controller',
                     'resourcequota': 'resource_quota',
                     'serviceaccount': 'service_account',
                     'statefulset': 'stateful_set'}
kubectl = SysCmdRunner('kubectl').run


class ClusterError(BatCaveException):
    """Kubernetes Cluster Exceptions.

    Attributes:
        BAD_ARGS: Bad arguments were specified.
        TIMEOUT: There was a timeout on the cluster.
    """
    BAD_ARGS = BatCaveError(1, Template('$error'))
    TIMEOUT = BatCaveError(2, Template('Timeout waiting for $seconds seconds for $what $action'))


class PodError(BatCaveException):
    """Kubernetes Pod Exceptions.

    Attributes:
        BAD_COPY_FILENAME: The file specified for the copy was invalid.
        COPY_ERROR: There was an error transferring a file to or from a pod.
        EXEC_ERROR: There was an error executing a pod command.
        FILE_NOT_FOUND: The file specified for the copy was not found in the pod.
        INVALID_COPY_MODE: The copy mode was invalid.
    """
    BAD_COPY_FILENAME = BatCaveError(1, Template('File not copied $mode, check filenames'))
    COPY_ERROR = BatCaveError(2, Template('Error copying pod file: $error'))
    EXEC_ERROR = BatCaveError(3, Template('Error executing pod command: $error'))
    FILE_NOT_FOUND = BatCaveError(4, Template('File not found in pod: $filename'))
    INVALID_COPY_MODE = BatCaveError(5, Template('Invalid pod file copy mode ($mode). Must be one of: (in, out)'))


class Cluster:
    """Class to create a universal abstract interface for a Kubernetes cluster."""

    def __init__(self, cluster_config: Optional[PathName] = None, context: Optional[str] = None):
        """
        Args:
            cluster_config (optional, default=None): The cluster configuration file to use.
            context (optional, default=None): The cluster configuration file context to use.

        Attributes:
            _batch_api: A reference to the BatchV1Api object.
            _config: The value of the cluster_config argument.
            _context: The value of the context argument.
            _core_api: A reference to the CoreV1Api object.
        """
        self._config = str(cluster_config) if isinstance(cluster_config, Path) else cluster_config
        self._context = context
        k8s_config.load_kube_config(self.config, self._context)
        self._core_api = CoreV1Api()
        self._apps_api = AppsV1Api()
        self._batch_api = BatchV1Api()

    def __getattr__(self, attr: str) -> Any:
        if '_' in attr:
            (verb, item_class_name) = attr.split('_', 1)
            if item_class_name.endswith('s'):
                item_class_name = item_class_name[0:-1]
                plural = 's'
            else:
                plural = ''
            item_class = globals()[item_class_name.capitalize()]
            method = getattr(self, f'{verb}_item{plural}')
            return lambda *a, **k: method(item_class, *a, **k)
        try:
            return getattr(self, f'get_{attr}')()
        except KeyError as err:
            raise AttributeError(f'No attribute for cluster: {attr}') from err

    config = property(lambda s: s._config, doc='A read-only property which returns configuration file for the cluster.')
    pod_exec = property(lambda s: s._core_api.connect_get_namespaced_pod_exec, doc='A read-only property which returns the pd exec function for the cluster.')

    def create_namespace(self, name: str, /, *, exists_ok: bool = False) -> 'Namespace':
        """Create a namespace.

        Args:
            name: The name of the namespace to create.
            exists_ok (optional, default=False): If True and the item already exists, delete before creating.

        Returns:
            The created namespace.
        """
        if not (self.has_item(Namespace, name) and exists_ok):
            namespace = V1Namespace()
            namespace.metadata = V1ObjectMeta()
            namespace.metadata.name = name
            self._core_api.create_namespace(namespace)
        return self.get_item(Namespace, name)

    def create_item(self, item_class: Type[K8sObject], item_spec: PathName, /, namespace: str = 'default', *, exists_ok: bool = False) -> K8sObject:
        """Create a new item using the specified spec.

        Args:
            item_class: The class of the item to create.
            item_spec: A file containing the Kubernetes specification.
            namespace (optional, default='default'): The Kubernetes namespace in which to create the item.
            exists_ok (optional, default=False): If True and the item already exists, delete before creating.

        Returns:
            The created item.
        """
        with open(item_spec, encoding=DEFAULT_ENCODING) as yaml_file:
            item_spec_content = yaml_load(yaml_file)
            item_name = item_spec_content['metadata']['name']
            if self.has_item(item_class, item_name, namespace=namespace) and exists_ok:
                self.find_method(item_class, 'delete')(item_name, namespace)
            return item_class(self, self.find_method(item_class, 'create')(namespace, item_spec_content))

    def create_job(self, job_spec: PathName, /, namespace: str = 'default', *, exists_ok: bool = False,
                   wait_for: bool = False, check_every: int = 2, timeout: bool = False) -> 'Job':
        """Create a job and wait for the specified condition.

        Args:
            job_spec: A file containing the Kubernetes job specification.
            namespace (optional, default='default'): The Kubernetes namespace in which to create the job.
            exists_ok (optional, default=False): If True and the item already exists, delete before creating.
            wait_for (optional, default=False): If True, wait until the job completes before returning.
            check_every (optional, default=2): The number of seconds to wait between every check to see if the job has completed.
            timeout (optional, default=False): If not False, this is the maximum number of seconds to wait for the job to complete.

        Raises:
            ClusterError.TIMEOUT: If timeout is True and the maximum number of seconds is exceeded.

        Returns:
            The created job.
        """
        if wait_for not in (False, 'start', 'finish'):
            raise ClusterError(ClusterError.BAD_ARGS, error=f"wait_for must be one of (start, finish, False) not '{wait_for}'")
        job = self.create_item(Job, job_spec, namespace=namespace, exists_ok=exists_ok)
        if not wait_for:
            return job

        start_time = dt.now()
        while not (job.status.active or job.status.succeeded or job.status.failed):
            if timeout and (dt.now() - start_time) > timedelta(seconds=timeout):
                raise ClusterError(ClusterError.TIMEOUT, seconds=timeout, what='Job', action='start')
            sleep(check_every)
            job = self.get_job(job.name, namespace)
        if wait_for == 'start':
            return job

        start_time = dt.now()
        while job.status.active:
            if timeout and (dt.now() - start_time) > timedelta(seconds=timeout):
                raise ClusterError(ClusterError.TIMEOUT, seconds=timeout, what='Job', action='finish')
            sleep(check_every)
            job = self.get_job(job.name, namespace)
        return job

    def delete_item(self, item_class: Type[K8sObject], name: str, /, namespace: str = 'default') -> None:
        """Delete the named item.

        Args:
            item_class: The class of the item to create.
            name: The name of the item to delete.
            namespace (optional, default='default'): The Kubernetes namespace from which to return the item.

        Returns:
            Nothing.
        """
        args = [name]
        if item_class.NAMESPACED:
            args.append(namespace)
        item_class(self, self.find_method(item_class, 'delete')(*args))

    def find_method(self, item_class: Type[K8sObject], method: str, /, suffix: str = '') -> Callable:
        """Search all the APIs for the specified method.

        Args:
            item_class: The item class to search.
            method: The method for which to search.
            suffix(optional, default=None): If not None, append to the method name with an underscore when searching.

        Returns:
            A reference to method.

        Raises:
            AttributeError: If the method is not found.
        """
        method_name = method
        if item_class.NAMESPACED:
            method_name += '_namespaced'
        k8s_class_name = XFORM_CLASS_NAMES.get(item_class_name := item_class.__name__.lower(), item_class_name)
        method_name += f'_{k8s_class_name}'
        if suffix:
            method_name += f'_{suffix}'
        for api in (self._core_api, self._apps_api, self._batch_api):
            if hasattr(api, method_name):
                return getattr(api, method_name)
        raise AttributeError(f'No method found: {method_name}')

    def get_item(self, item_class: Type[K8sObject], name: str, /, namespace: str = 'default') -> K8sObject:
        """Get the requested item.

        Args:
            item_class: The item class.
            item_name: The name of the item to return.
            namespace (optional, default='default'): The Kubernetes namespace from which to return the item.

        Returns:
            The requested item.
        """
        args = [name]
        if item_class.NAMESPACED:
            args.append(namespace)
        return item_class(self, self.find_method(item_class, 'read')(*args))

    def get_items(self, item_class: Type[K8sObject], /, namespace: str = 'default', **keys) -> List[K8sObject]:
        """Get all the item of the requested type.

        Args:
            item_class: The item class.
            namespace (optional, default='default'): The Kubernetes namespace from which to return the items.
            keys (optional): A list of keys by which to filter the items.

        Returns:
            The requested item list.
        """
        if item_class.NAMESPACED:
            keys['namespace'] = namespace
        return [item_class(self, i) for i in self.find_method(item_class, 'list')(**keys).items]

    def has_item(self, item_class: Type[K8sObject], item_name: str, /, namespace: str = 'default') -> bool:
        """Determine if the named items of the specified class exists.

        Args:
            item_class: The item class for which to search.
            item_name: The name of the item to search for.
            namespace (optional, default='default'): The Kubernetes namespace to search.

        Returns:
            Returns True if the named item exists, False otherwise.
        """
        return bool([i for i in self.get_items(item_class, namespace=namespace) if i.name == item_name])

    def kubectl(self, *args, **kwargs) -> CommandResult:
        """Run a kubectl command.

        Args:
            *args: The arguments to pass to kubectl.
            *kwargs: The named arguments to pass to kubectl.

        Returns:
            The result of the kubectl command.
        """
        config_args = {}
        if self.config:
            config_args['kubeconfig'] = self.config
        if self._context:
            config_args['context'] = self._context
        return kubectl(*args, **config_args, **kwargs)

    def patch_item(self, item_class: Type[K8sObject], name: str, item_spec: K8sObject, /, namespace: str = 'default') -> K8sObject:
        """Patch the named item.

        Args:
            item_class: The class of the item to patch.
            name: The name of the item to patch.
            item_spec: The new specification for the patched item.
            namespace (optional, default='default'): The Kubernetes namespace from which to return the item.

        Returns:
            The patched item.
        """
        args: list[Any] = [name]
        if item_class.NAMESPACED:
            args.append(namespace)
        args.append(item_spec.api_object)
        return item_class(self, self.find_method(item_class, 'patch')(*args))


class ClusterObject:  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a Kubernetes cluster object.

    Attributes:
        NAMESPACED: If True, the object is a cluster namespaced object.
    """
    NAMESPACED = True

    def __init__(self, cluster: Cluster, object_ref: Any, /):
        """
        Args:
            cluster: The cluster containing this object.
            object_ref: A reference to the underlying API object.

        Attributes:
            _cluster_obj: The value of the cluster argument.
            _object_ref: The value of the object_ref argument.
        """
        self._cluster_obj = cluster
        self._object_ref = object_ref

    def __getattr__(self, attr: str) -> 'ClusterObject':
        if hasattr(self._object_ref, attr):
            return getattr(self._object_ref, attr)
        return getattr(self._object_ref.metadata, attr)

    api_object = property(lambda s: s._object_ref, doc='A read-only property which returns Kubernetes API object.')


class Binding(ClusterObject):  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a Kubernetes cron job."""


class Configmap(ClusterObject):  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a Kubernetes cron job."""


class Cronjob(ClusterObject):  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a Kubernetes cron job."""


class Daemonset(ClusterObject):  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a Kubernetes cron job."""


class Deployment(ClusterObject):  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a Kubernetes deployment."""


class Endpoints(ClusterObject):  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a Kubernetes deployment."""


class Event(ClusterObject):  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a Kubernetes deployment."""


class Job(ClusterObject):  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a Kubernetes job."""


class Limitrange(ClusterObject):  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a Kubernetes deployment."""


class Namespace(ClusterObject):  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a Kubernetes namespace."""
    NAMESPACED = False


class Persistentvolumeclaim(ClusterObject):  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a Kubernetes deployment."""


class Pod(ClusterObject):
    """Class to create a universal abstract interface for a Kubernetes pod."""

    logs = property(lambda s: s._cluster_obj.kubectl('logs', s.name, namespace=s.namespace), doc='A read-only property which returns the pod logs.')

    def cp_file(self, mode: str, source: PathName, target: PathName, /) -> None:
        """Copy a file into or out of the pod.

        Args:
            mode: The direction of the copy ('in' or 'out').
            source: The source file for the copy.
            target: The target file for the copy.

        Returns:
            Nothing.

        Raises:
            PodError.BAD_COPY_FILENAME: If the source or target name was not found.
            PodError.COPY_ERROR: If there was an error when copying the file.
            PodError.INVALID_COPY_MODE: If the specified mode is not known.
        """
        source_path: PathName
        target_path: PathName
        if mode == 'in':
            source_path = Path(source)
            target_path = f'{self.namespace}/{self.name}:{target}'
        elif mode == 'out':
            source_path = f'{self.namespace}/{self.name}:{source}'
            target_path = Path(target)
        else:
            raise PodError(PodError.INVALID_COPY_MODE, mode=mode)

        output = self._cluster_obj.kubectl('cp', str(source_path), str(target_path))
        if not (self.has_file(str(target)) if (mode == 'in') else cast(Path, target_path).exists()):  # pylint: disable=superfluous-parens
            if output:
                raise PodError(PodError.COPY_ERROR, error=output)
            raise PodError(PodError.BAD_COPY_FILENAME, mode=mode)

    def exec(self, *command, **k8s_api_kwargs) -> str:
        """Execute a command in the pod.

        Args:
            *command: The command and arguments to execute.
            **k8s_api_kwargs: The Kubernetes API parameters to pass to the stream call.

        Returns:
            The output from the command.

        Raises:
            PodError.EXEC_ERROR: If the word 'error' occurs in the output.
        """
        output = k8s_process(self._cluster_obj.pod_exec, self.name, self.namespace,
                             command=list(command), stderr=True, stdin=False, stdout=True, tty=False, _preload_content=True, **k8s_api_kwargs)
        if 'error' in output:
            raise PodError(PodError.EXEC_ERROR, error=output)
        return output.split('\n')[0:-1]  # TODO: This will not work when run on Linux against a Windows container

    def get_file(self, source: str, target: Optional[PathName] = None, /) -> None:
        """Copy a file out of the pod.

        Args:
            source: The source for the copy.
            target (optional, default=None): If not None, the target for the copy.

        Returns:
            Nothing.
        """
        target_path = Path(target) if target else Path(Path(source).name)
        self.cp_file('out', source, target_path)

    def has_file(self, filename: str, /) -> bool:
        """Determine if the pod has the specified file.

        Args:
            filename: The name of the file for which to search.

        Returns:
            True if the specified file exists, False otherwise.
        """
        return filename == self.exec('ls', filename)[0]

    def put_file(self, source: PathName, target: str, /) -> None:
        """Copy a file into the pod.

        Args:
            source: The source for the copy.
            target: The target for the copy.

        Returns:
            Nothing.
        """
        self.cp_file('in', source, target)

    def remove_file(self, filename: str, /, *, not_exists_ok: bool = False) -> None:
        """Remove the specified file from the pod.

        Args:
            filename: The name of the file to remove.
            not_exists_ok (optional, default=False): If True, raise an exception if the file does not exist.

        Returns:
            Nothing.

        Raises:
            PodError.FILE_NOT_FOUND: If the file is not found.
        """
        if not (not_exists_ok or self.has_file(filename)):
            raise PodError(PodError.FILE_NOT_FOUND, filename=filename)
        self.exec('rm', filename)


class Replicationcontroller(ClusterObject):  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a Kubernetes service."""


class Resourcequota(ClusterObject):  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a Kubernetes service."""


class Secret(ClusterObject):  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a Kubernetes service."""


class Service(ClusterObject):  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a Kubernetes service."""


class Serviceaccount(ClusterObject):  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a Kubernetes service."""


class Replicaset(ClusterObject):  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a Kubernetes service."""


class Statefulset(ClusterObject):  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a Kubernetes service."""

# cSpell:ignore kube kubeconfig batcave cronjob xform configmap limitrange persistentvolumeclaim replicationcontroller resourcequota serviceaccount daemonset replicaset statefulset



"""This module provides utilities for creating reports.

Examples:
    The simplest usage would be::

        report = reporter.Report('REPORT HEADER', 'REPORT FOOTER')

    To add a line to the report::

        line = reporter.Line('a line')
        report.add_line(line)

    To add a section::

        section1 = reporter.Section('Section 1 Header', 'Section 1 Footer')
        report.add_section(section1)

    To add a link::

        link = reporter.Link('Link Text', 'URL')
        report.register_link(link)
        report.add_line(reporter.Line(f'Line with a {link} in it'))

    To add a list of links::

        links = reporter.LinkList({'link1': 'http://link1', 'link2': 'http://link2'})
        report.register_link(links)
        report.add_line(reporter.Line(f'Line with {links}s in it'))

    A table is a list of lists with each outer list a row

    To add a table::

        rows = [['11', '12'], ['21', '22']]
        table = reporter.Table(rows, 'Table Header', 'Table Footer')
        report.add_table(table)

    To embed a link in a table::

        link = reporter.Link('Table Text', 'http://table')
        report.register_link(link)
        rows = [['11', '12'], ['21', link]]
        table = reporter.Table(rows, 'Table Header', 'Table Footer')
        report.add_table(table)

    A simple report can be output with::

        print(report)

Attributes:
    Each attribute is made up of what it affects and where that effect takes place.
        i.e. PART_PIECE_WHERE

    Part abbreviations:
        rpt = report
        sec = section
        tbl = table
        lin = line
        lnk = link
        lst = list
    Piece abbreviations:
        hdr = header
        bdy = body
        row = DUH!
        cel = cell
        ftr = footer
    Where abbreviations:
        ldr = leader
        int = interstitial
        trm = terminator

    The report/section structure is
    sec_ldr - header - body - footer - sec_trm
    where
    header = sec_hdr_ldr - HEADER - sec_hdr_trm
    body = sec_bdy_ldr - BODY - sec_bdy_trm
    footer = sec_ftr_ldr - FOOTER - sec_ftr_trm

    The table structure is
    tbl_ldr - header - body - footer - tbl_trm
    where the body is
    tbl_bdy_ldr - rows - tbl_bdr_trm
    and rows are
    tbl_row_ldr - cell - tbl_row_trm
    and cells are
    tbl_cel_ldr - data - tbl_cel_trm

    The general line structure is:
    lin_ldr - LINE - lin_trm

    The general list structure is:
    lst_ldr - ITEM - lst_int ITEM - lst_trm
"""

# Import standard modules
from copy import deepcopy
from enum import Enum
from typing import cast, Dict, List, Optional, Type

LIN_LDR_ATTR = 'lin_ldr'
LIN_TRM_ATTR = 'lin_trm'
LNK_LDR_ATTR = 'lnk_ldr'
LNK_TRM_ATTR = 'lnk_trm'
LST_INT_ATTR = 'lst_int'
LST_LDR_ATTR = 'lst_ldr'
LST_TRM_ATTR = 'lst_trm'
OUTPUT_ATTR = 'output'
RPT_BDY_LDR_ATTR = 'rpt_bdy_ldr'
RPT_BDY_TRM_ATTR = 'rpt_bdy_trm'
RPT_FTR_LDR_ATTR = 'rpt_ftr_ldr'
RPT_FTR_TRM_ATTR = 'rpt_ftr_trm'
RPT_HDR_LDR_ATTR = 'rpt_hdr_ldr'
RPT_HDR_TRM_ATTR = 'rpt_hdr_trm'
RPT_LDR_ATTR = 'rpt_ldr'
RPT_TRM_ATTR = 'rpt_trm'
SEC_BDY_LDR_ATTR = 'sec_bdy_ldr'
SEC_BDY_TRM_ATTR = 'sec_bdy_trm'
SEC_FTR_LDR_ATTR = 'sec_ftr_ldr'
SEC_FTR_TRM_ATTR = 'sec_ftr_trm'
SEC_HDR_LDR_ATTR = 'sec_hdr_ldr'
SEC_HDR_TRM_ATTR = 'sec_hdr_trm'
SEC_LDR_ATTR = 'sec_ldr'
SEC_TRM_ATTR = 'sec_trm'
TBL_BDY_LDR_ATTR = 'tbl_bdy_ldr'
TBL_BDY_TRM_ATTR = 'tbl_bdy_trm'
TBL_CEL_LDR_ATTR = 'tbl_cel_ldr'
TBL_CEL_TRM_ATTR = 'tbl_cel_trm'
TBL_FTR_LDR_ATTR = 'tbl_ftr_ldr'
TBL_FTR_TRM_ATTR = 'tbl_ftr_trm'
TBL_HDR_LDR_ATTR = 'tbl_hdr_ldr'
TBL_HDR_TRM_ATTR = 'tbl_hdr_trm'
TBL_LDR_ATTR = 'tbl_ldr'
TBL_TRM_ATTR = 'tbl_trm'
TBL_ROW_LDR_ATTR = 'tbl_row_ldr'
TBL_ROW_TRM_ATTR = 'tbl_row_trm'


class SimpleAttribute:
    """Class to create a universal abstract interface for a report attribute which has a default value and a list of valid values."""

    def __init__(self, default: str, /, *other):
        """
        Args:
            default: The default value of the attribute.
            other (optional): a list of other allowed values.

        Attributes:
            _valid: The value of the other argument with default appended to the list.
            _value: The value of the default argument.
        """
        self._value = default
        self._valid = list(other)
        self._valid.append(default)

    count = property(lambda s: len(s._valid), doc='A read-only property which returns the number of valid attribute values.')

    @property
    def value(self) -> str:
        """A read-write property which returns and sets the value of the attribute."""
        return self._value

    @value.setter
    def value(self, val: str, /) -> None:
        if not self.is_valid(val):
            raise ValueError('invalid value for SimpleAttribute: ' + val)
        self._value = val

    def is_valid(self, attr: str, /) -> bool:
        """Determine if the specified attribute is valid.

        Args:
            attr: The attribute to validate.

        Returns:
            True if the attribute it valid, False otherwise.
        """
        return attr in self._valid


class MetaAttribute:  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a report attribute which returns a value based on the value of a SimpleAttribute."""

    def __init__(self, attr: str, /, **val_map):
        """
        Args:
            attr: The attribute.
            val_map (optional): a dictionary of other allowed values.

        Attributes:
            _attr: The value of the attr argument.
            _value_map: The value of the val_map argument.
        """
        self._attr = attr
        self._value_map = val_map

    def _set_values(self, val_map: Dict[str, str], /) -> None:
        """Set the value of a collection of attributes.

        Args:
            val_map: The collection of attributes to set.

        Returns:
            Nothing.
        """
        self._value_map = val_map

    simple_attr_name = property(lambda s: s._attr, doc='A read-only property which returns the simple attribute name.')
    values = property(fset=_set_values, doc='A read-only property which returns the values of the attribute as a dictionary.')

    def get_value(self, attr: str, /) -> str:
        """Get the value of an attribute.

        Args:
            attr: The attribute for which to return the value.

        Returns:
            The value of the attribute.
        """
        return self._value_map[attr]


Attribute = SimpleAttribute | MetaAttribute

_ATTRIBUTES = {OUTPUT_ATTR: SimpleAttribute('html', 'text'),
               RPT_LDR_ATTR: MetaAttribute(OUTPUT_ATTR, text='\n', html='<html><meta http-equiv="Content-Type" content="text/html;charset=utf-8"><body><center>'),
               RPT_TRM_ATTR: MetaAttribute(OUTPUT_ATTR, text='\n', html='</center></body></html>'),
               RPT_HDR_LDR_ATTR: MetaAttribute(OUTPUT_ATTR, text=('-' * 79) + '\n', html='<h1>'),
               RPT_HDR_TRM_ATTR: MetaAttribute(OUTPUT_ATTR, text=('-' * 79) + '\n', html='</h1>'),
               RPT_BDY_LDR_ATTR: MetaAttribute(OUTPUT_ATTR, text='', html=''),
               RPT_BDY_TRM_ATTR: MetaAttribute(OUTPUT_ATTR, text='', html=''),
               RPT_FTR_LDR_ATTR: MetaAttribute(OUTPUT_ATTR, text=('-' * 79) + '\n', html=''),
               RPT_FTR_TRM_ATTR: MetaAttribute(OUTPUT_ATTR, text=('-' * 79) + '\n', html=''),
               SEC_LDR_ATTR: MetaAttribute(OUTPUT_ATTR, text='', html=''),
               SEC_TRM_ATTR: MetaAttribute(OUTPUT_ATTR, text=('=' * 79) + '\n', html=''),
               SEC_HDR_LDR_ATTR: MetaAttribute(OUTPUT_ATTR, text='', html='<h2>'),
               SEC_HDR_TRM_ATTR: MetaAttribute(OUTPUT_ATTR, text='', html='</h2>'),
               SEC_BDY_LDR_ATTR: MetaAttribute(OUTPUT_ATTR, text='', html=''),
               SEC_BDY_TRM_ATTR: MetaAttribute(OUTPUT_ATTR, text='', html=''),
               SEC_FTR_LDR_ATTR: MetaAttribute(OUTPUT_ATTR, text='', html=''),
               SEC_FTR_TRM_ATTR: MetaAttribute(OUTPUT_ATTR, text='', html=''),
               TBL_LDR_ATTR: MetaAttribute(OUTPUT_ATTR, text='', html='<table border="1">'),
               TBL_TRM_ATTR: MetaAttribute(OUTPUT_ATTR, text='', html='</table><br>'),
               TBL_HDR_LDR_ATTR: MetaAttribute(OUTPUT_ATTR, text='', html='<td colspan="2" align="center"><h2>'),
               TBL_HDR_TRM_ATTR: MetaAttribute(OUTPUT_ATTR, text='\n', html='</h2></td>'),
               TBL_BDY_LDR_ATTR: MetaAttribute(OUTPUT_ATTR, text='', html=''),
               TBL_BDY_TRM_ATTR: MetaAttribute(OUTPUT_ATTR, text='', html=''),
               TBL_ROW_LDR_ATTR: MetaAttribute(OUTPUT_ATTR, text='|', html='<tr>'),
               TBL_ROW_TRM_ATTR: MetaAttribute(OUTPUT_ATTR, text='\n', html='</tr>'),
               TBL_CEL_LDR_ATTR: MetaAttribute(OUTPUT_ATTR, text=' ', html='<td>'),
               TBL_CEL_TRM_ATTR: MetaAttribute(OUTPUT_ATTR, text=' |', html='</td>'),
               TBL_FTR_LDR_ATTR: MetaAttribute(OUTPUT_ATTR, text='', html='<td colspan="2" align="center"><h2>'),
               TBL_FTR_TRM_ATTR: MetaAttribute(OUTPUT_ATTR, text='\n', html='</h2></td>'),
               LIN_LDR_ATTR: MetaAttribute(OUTPUT_ATTR, text='', html=''),
               LIN_TRM_ATTR: MetaAttribute(OUTPUT_ATTR, text='\n', html='<br>'),
               LNK_LDR_ATTR: MetaAttribute(OUTPUT_ATTR, text='', html='<a href="%s">'),
               LNK_TRM_ATTR: MetaAttribute(OUTPUT_ATTR, text='', html='</a>'),
               LST_LDR_ATTR: MetaAttribute(OUTPUT_ATTR, text='', html=''),
               LST_INT_ATTR: MetaAttribute(OUTPUT_ATTR, text=', ', html=', '),
               LST_TRM_ATTR: MetaAttribute(OUTPUT_ATTR, text='', html='')}


class ReportObject:  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a report object."""

    def __init__(self, container: Optional[Type['ReportObject']] = None, /, **attributes):
        """
        Args:
            container: The container for this object.
            **attributes (optional): A dictionary of attributes for the object.

        Attributes:
            container: The value of the container argument.
            _attributes: A dictionary of attributes for this object as initialized by the attr argument.
        """
        self._attributes: Dict[str, Attribute] = {}
        self.container: Optional[Type[ReportObject]] = container
        for (attr, val) in attributes.items():
            self._set_attribute(attr, val)

    def _get_attr_ref(self, attr: str, /) -> Attribute:
        """Get a reference to the requested attributes.

        Args:
            attr: The attribute for which to return the reference.

        Returns:
            A reference to the requested attribute.

        Raises:
            AttributeError: If the requested attribute is not found.
        """
        if attr in self._attributes:
            return self._attributes[attr]
        if self.container:
            return cast('ReportObject', self.container)._get_attr_ref(attr)  # pylint: disable=protected-access
        if attr in _ATTRIBUTES:
            return cast(Attribute, _ATTRIBUTES[attr])
        raise AttributeError(f"'{type(self)}' object has no attribute '{attr}'")

    def _get_attribute(self, attr: str, /) -> str:
        """Get the value of the requested attributes.

        Args:
            attr: The attribute for which to return the value.

        Returns:
            A value of the requested attribute.
        """
        if isinstance((attr_ref := self._get_attr_ref(attr)), MetaAttribute):
            sub_attr_ref = cast(SimpleAttribute, self._get_attr_ref(attr_ref.simple_attr_name))
            return attr_ref.get_value(sub_attr_ref.value)
        return attr_ref.value

    def _set_attribute(self, attr: str, val: str, /) -> None:
        """Set the value of the requested attributes.

        Args:
            attr: The attribute for which to set the value.
            val: The value to which to set the attribute.

        Returns:
            Nothing.
        """
        if attr not in self._attributes:
            self._attributes[attr] = deepcopy(self._get_attr_ref(attr))
            if isinstance(self._attributes[attr], MetaAttribute):
                cast(MetaAttribute, self._attributes[attr]).values = val
            else:
                cast(SimpleAttribute, self._attributes[attr]).value = val

    lin_ldr = property(lambda s: s._get_attribute(LIN_LDR_ATTR), lambda s, v: s._set_attribute(LIN_LDR_ATTR, v), doc='A read-write property for the line leader attribute.')
    lin_trm = property(lambda s: s._get_attribute(LIN_TRM_ATTR), lambda s, v: s._set_attribute(LIN_TRM_ATTR, v), doc='A read-write property for the line terminator attribute.')
    lnk_ldr = property(lambda s: s._get_attribute(LNK_LDR_ATTR), lambda s, v: s._set_attribute(LNK_LDR_ATTR, v), doc='A read-write property for the link leader attribute.')
    lnk_trm = property(lambda s: s._get_attribute(LNK_TRM_ATTR), lambda s, v: s._set_attribute(LNK_TRM_ATTR, v), doc='A read-write property for the link terminator attribute.')
    lst_int = property(lambda s: s._get_attribute(LST_INT_ATTR), lambda s, v: s._set_attribute(LST_INT_ATTR, v), doc='A read-write property for the list separator attribute.')
    lst_ldr = property(lambda s: s._get_attribute(LST_LDR_ATTR), lambda s, v: s._set_attribute(LST_LDR_ATTR, v), doc='A read-write property for the list leader attribute.')
    lst_trm = property(lambda s: s._get_attribute(LST_TRM_ATTR), lambda s, v: s._set_attribute(LST_TRM_ATTR, v), doc='A read-write property for the list terminator attribute.')
    output = property(lambda s: s._get_attribute(OUTPUT_ATTR), lambda s, v: s._set_attribute(OUTPUT_ATTR, v), doc='A read-write property for the output attribute.')
    rpt_bdy_ldr = property(lambda s: s._get_attribute(RPT_BDY_LDR_ATTR), lambda s, v: s._set_attribute(RPT_BDY_LDR_ATTR, v), doc='A read-write property for the report body leader attribute.')
    rpt_bdy_trm = property(lambda s: s._get_attribute(RPT_BDY_TRM_ATTR), lambda s, v: s._set_attribute(RPT_BDY_TRM_ATTR, v), doc='A read-write property for the report body terminator attribute.')
    rpt_ftr_ldr = property(lambda s: s._get_attribute(RPT_FTR_LDR_ATTR), lambda s, v: s._set_attribute(RPT_FTR_LDR_ATTR, v), doc='A read-write property for the report footer leader attribute.')
    rpt_ftr_trm = property(lambda s: s._get_attribute(RPT_FTR_TRM_ATTR), lambda s, v: s._set_attribute(RPT_FTR_TRM_ATTR, v), doc='A read-write property for the report footer terminator attribute.')
    rpt_hdr_ldr = property(lambda s: s._get_attribute(RPT_HDR_LDR_ATTR), lambda s, v: s._set_attribute(RPT_HDR_LDR_ATTR, v), doc='A read-write property for the report header leader attribute.')
    rpt_hdr_trm = property(lambda s: s._get_attribute(RPT_HDR_TRM_ATTR), lambda s, v: s._set_attribute(RPT_HDR_TRM_ATTR, v), doc='A read-write property for the report header terminator attribute.')
    rpt_ldr = property(lambda s: s._get_attribute(RPT_LDR_ATTR), lambda s, v: s._set_attribute(RPT_LDR_ATTR, v), doc='A read-write property for the report leader attribute.')
    rpt_trm = property(lambda s: s._get_attribute(RPT_TRM_ATTR), lambda s, v: s._set_attribute(RPT_TRM_ATTR, v), doc='A read-write property for the report terminator attribute.')
    sec_bdy_ldr = property(lambda s: s._get_attribute(SEC_BDY_LDR_ATTR), lambda s, v: s._set_attribute(SEC_BDY_LDR_ATTR, v), doc='A read-write property for the section body leader attribute.')
    sec_bdy_trm = property(lambda s: s._get_attribute(SEC_BDY_TRM_ATTR), lambda s, v: s._set_attribute(SEC_BDY_TRM_ATTR, v), doc='A read-write property for the section body terminator attribute.')
    sec_ftr_ldr = property(lambda s: s._get_attribute(SEC_FTR_LDR_ATTR), lambda s, v: s._set_attribute(SEC_FTR_LDR_ATTR, v), doc='A read-write property for the section footer leader attribute.')
    sec_ftr_trm = property(lambda s: s._get_attribute(SEC_FTR_TRM_ATTR), lambda s, v: s._set_attribute(SEC_FTR_TRM_ATTR, v), doc='A read-write property for the section footer terminator attribute.')
    sec_hdr_ldr = property(lambda s: s._get_attribute(SEC_HDR_LDR_ATTR), lambda s, v: s._set_attribute(SEC_HDR_LDR_ATTR, v), doc='A read-write property for the section header leader attribute.')
    sec_hdr_trm = property(lambda s: s._get_attribute(SEC_HDR_TRM_ATTR), lambda s, v: s._set_attribute(SEC_HDR_TRM_ATTR, v), doc='A read-write property for the section header terminator attribute.')
    sec_ldr = property(lambda s: s._get_attribute(SEC_LDR_ATTR), lambda s, v: s._set_attribute(SEC_LDR_ATTR, v), doc='A read-write property for the section leader attribute.')
    sec_trm = property(lambda s: s._get_attribute(SEC_TRM_ATTR), lambda s, v: s._set_attribute(SEC_TRM_ATTR, v), doc='A read-write property for the section terminator attribute.')
    tbl_bdy_ldr = property(lambda s: s._get_attribute(TBL_BDY_LDR_ATTR), lambda s, v: s._set_attribute(TBL_BDY_LDR_ATTR, v), doc='A read-write property for the table body leader attribute.')
    tbl_bdy_trm = property(lambda s: s._get_attribute(TBL_BDY_TRM_ATTR), lambda s, v: s._set_attribute(TBL_BDY_TRM_ATTR, v), doc='A read-write property for the table body terminator attribute.')
    tbl_cel_ldr = property(lambda s: s._get_attribute(TBL_CEL_LDR_ATTR), lambda s, v: s._set_attribute(TBL_CEL_LDR_ATTR, v), doc='A read-write property for the table cell leader attribute.')
    tbl_cel_trm = property(lambda s: s._get_attribute(TBL_CEL_TRM_ATTR), lambda s, v: s._set_attribute(TBL_CEL_TRM_ATTR, v), doc='A read-write property for the table cell terminator attribute.')
    tbl_ftr_ldr = property(lambda s: s._get_attribute(TBL_FTR_LDR_ATTR), lambda s, v: s._set_attribute(TBL_FTR_LDR_ATTR, v), doc='A read-write property for the table footer leader attribute.')
    tbl_ftr_trm = property(lambda s: s._get_attribute(TBL_FTR_TRM_ATTR), lambda s, v: s._set_attribute(TBL_FTR_TRM_ATTR, v), doc='A read-write property for the table footer terminator attribute.')
    tbl_hdr_ldr = property(lambda s: s._get_attribute(TBL_HDR_LDR_ATTR), lambda s, v: s._set_attribute(TBL_HDR_LDR_ATTR, v), doc='A read-write property for the table header leader attribute.')
    tbl_hdr_trm = property(lambda s: s._get_attribute(TBL_HDR_TRM_ATTR), lambda s, v: s._set_attribute(TBL_HDR_TRM_ATTR, v), doc='A read-write property for the table header terminator attribute.')
    tbl_ldr = property(lambda s: s._get_attribute(TBL_LDR_ATTR), lambda s, v: s._set_attribute(TBL_LDR_ATTR, v), doc='A read-write property for the table leader attribute.')
    tbl_trm = property(lambda s: s._get_attribute(TBL_TRM_ATTR), lambda s, v: s._set_attribute(TBL_TRM_ATTR, v), doc='A read-write property for the table terminator attribute.')
    tbl_row_ldr = property(lambda s: s._get_attribute(TBL_ROW_LDR_ATTR), lambda s, v: s._set_attribute(TBL_ROW_LDR_ATTR, v), doc='A read-write property for the table row leader attribute.')
    tbl_row_trm = property(lambda s: s._get_attribute(TBL_ROW_TRM_ATTR), lambda s, v: s._set_attribute(TBL_ROW_TRM_ATTR, v), doc='A read-write property for the table row terminator attribute.')

    @property
    def depth(self) -> int:
        """A read-only property which returns the report depth of this object."""
        if self.container:
            return cast(ReportObject, self.container).depth + 1
        return 1


class Section(ReportObject):
    """Class to create a universal abstract interface for a report section."""

    def __init__(self, /, *, header: str = '', footer: str = '', cont: Optional[Type[ReportObject]] = None, **attr):
        """
        Args:
            header (optional, default=''): The section header.
            footer (optional, default=''): The section footer.
            cont: The container for this section.
            attr (optional): A dictionary of attributes for the section.

        Attributes:
            footer: The value of the footer argument.
            header: The value of the header argument.
            _members: A list of objects contained in this section.
        """
        super().__init__(cont, **attr)
        self.header = header
        self.footer = footer
        self._members: List[Type[ReportObject]] = []

    def __str__(self):
        the_str = self.sec_ldr
        if self.header:
            the_str += self.sec_hdr_ldr + str(Line(self.header, self)) + self.sec_hdr_trm
        the_str += self.sec_bdy_ldr + ''.join([str(part) for part in self._members]) + self.sec_bdy_trm
        if self.footer:
            the_str += self.sec_ftr_ldr + str(Line(self.footer, self)) + self.sec_ftr_trm
        return the_str + self.sec_trm

    def add_line(self, line: Type['Line'], /) -> None:
        """Add a line to the section.

        Args:
            line: The line to add to the section.

        Returns:
            Nothing.
        """
        self.add_member(line)

    def add_member(self, thing: Type[ReportObject], /) -> None:
        """Add a member to the section.

        Args:
            thing: The member to add to the section.

        Returns:
            Nothing.
        """
        self._members.append(thing)
        thing.container = cast(Type[ReportObject], self)

    def add_section(self, section: Type['Section'], /) -> None:
        """Add a sub-section to the section.

        Args:
            section: The sub-section to add to the section.

        Returns:
            Nothing.
        """
        self.add_member(section)

    def add_table(self, table: Type['Table'], /) -> None:
        """Add a table to the section.

        Args:
            table: The table to add to the section.

        Returns:
            Nothing.
        """
        self.add_member(table)

    def register_link(self, link: 'Link', /) -> None:
        """Register a hyperlink in the section.

        Args:
            link: The hyperlink to register.

        Returns:
            Nothing.
        """
        link.container = cast(Type[ReportObject], self)


class Report(Section):
    """Class to create a universal abstract interface for a report."""

    def __str__(self):
        the_str = self.rpt_ldr
        if self.header:
            the_str += self.rpt_hdr_ldr + str(Line(self.header, self)) + self.rpt_hdr_trm
        the_str += self.rpt_bdy_ldr + ''.join([str(part) for part in self._members]) + self.rpt_bdy_trm
        if self.footer:
            the_str += self.rpt_ftr_ldr + str(Line(self.footer, self)) + self.rpt_ftr_trm
        return the_str + self.rpt_trm


class Cell(ReportObject):
    """Class to create a universal abstract interface for a cell in a table in a report."""

    def __init__(self, data: str, cont: Optional[Type[ReportObject]] = None, /, **attr):
        """
        Args:
            data: The cell data.
            cont (optional, default=None): The container for this cell.
            attr (optional): A dictionary of attributes for the cell.

        Attributes:
            _data: The value of the data argument.
        """
        super().__init__(cont, **attr)
        self._data = data

    def __str__(self):
        data_str = self._data
        if isinstance(self._data, (int, list, tuple, Enum, LinkList, Link)) or not self._data:
            data_str = str(self._data)
        return self.tbl_cel_ldr + data_str + self.tbl_cel_trm


class Table(ReportObject):
    """Class to create a universal abstract interface for a report section table."""

    def __init__(self, data: List[List[str]], /, *, header: str = '', footer: str = '', **attr):
        """
        Args:
            data: The table data.
            header (optional, default=''): The table header.
            footer (optional, default=''): The table footer.
            attr (optional): A dictionary of attributes for the table.

        Attributes:
            footer: The value of the footer argument.
            header: The value of the header argument.
            _data: The value of the data argument.
        """
        super().__init__(**attr)
        self.header = header
        self.footer = footer
        self._data = data

    def __str__(self):
        col_widths = []
        if self.output == 'text':
            for row in self._data:
                i = 0
                for col in row:
                    if (i + 1) > len(col_widths):
                        col_widths.append(0)
                    col_widths[i] = max(col_widths[i], len(col))
                    i += 1

        the_str = self.tbl_ldr
        if self.header:
            the_str += self.tbl_hdr_ldr + self.header + self.tbl_hdr_trm
        the_str += self.tbl_bdy_ldr
        for row in self._data:
            the_str += self.tbl_row_ldr
            i = 0
            for col in row:
                if self.output == 'text':
                    col_pad = ' ' * int((col_widths[i] - len(col)) / 2)
                    col_str = col_pad + col + col_pad
                    col_str += ' ' * (col_widths[i] - len(col_str))
                else:
                    col_str = col
                the_str += str(Cell(col_str, self))
                i += 1
            the_str += self.tbl_row_trm
        the_str += self.tbl_bdy_trm
        if self.footer:
            the_str += self.tbl_ftr_ldr + self.footer + self.tbl_ftr_trm
        return the_str + self.tbl_trm


class Line(ReportObject):
    """Class to create a universal abstract interface for a report section line."""

    def __init__(self, text: str, cont: Optional[Type[ReportObject]] = None, /, **attr):
        """
        Args:
            text: The line text.
            cont (optional, default=None): The container for this line.
            attr (optional): A dictionary of attributes for the line.

        Attributes:
            _text: The value of the text argument.
        """
        super().__init__(cont, **attr)
        self._text = text

    def __str__(self):
        return self.lin_ldr + self._text + self.lin_trm


class Link(Line):
    """Class to create a universal abstract interface for a report hyperlink."""

    def __init__(self, text: str, /, url: str = '', cont: Optional[Type[ReportObject]] = None, **attr):
        """
        Args:
            text: The link text.
            url (optional, default=None): The link URL.
            cont (optional, default=None): The container for this link.
            attr (optional): A dictionary of attributes for the link.

        Attributes:
            _url: The value of the url argument.
        """
        super().__init__(text, cont, **attr)
        self._url = url

    def __str__(self):
        try:
            url = self.lnk_ldr % self._url
        except TypeError as err:
            if 'not all arguments converted during string formatting' not in str(err):
                raise
            url = self.lnk_ldr
        return url + self._text + self.lnk_trm


class LinkList(Section):
    """Class to create a universal abstract interface for a list of hyperlinks in a report."""

    def __init__(self, urls: Dict[str, str], cont: Optional[Type[ReportObject]] = None, /, **attr):
        """
        Args:
            urls: The list of URLs the section.
            cont (optional, default=None): The container for this section.
            attr (optional): A dictionary of attributes for the section.

        Attributes:
            _list: The value of the urls argument converted into links.
        """
        super().__init__(cont=cont, **attr)
        self._list = []
        for key in sorted(urls.keys()):
            self.register_link(link := Link(key, urls[key]))
            self._list.append(link)

    def __str__(self):
        return self.lst_int.join([str(item) for item in self._list])



"""This module provides utilities for working with files.

Attributes:
    ConversionMode (Enum): The conversion modes for the eol_convert function.
    PACKER_CLASSES (dict): A mapping of file compression extensions to the classes that create them.
    COMPRESSION_TYPE (dict): A mapping of file compression extensions to compression types.
"""

# Import standard modules
from bz2 import BZ2File
from enum import Enum
from gzip import GzipFile
from logging import getLogger
from lzma import LZMAFile
from pathlib import Path
from shutil import copy
from stat import S_IRWXU
from string import Template
from tarfile import open as tar_open
from time import time
from typing import Any, Iterable, List, Optional, Tuple
from zipfile import ZipFile, ZIP_DEFLATED

# Import internal modules
from .sysutil import popd, pushd
from .lang import DEFAULT_ENCODING, BatCaveError, BatCaveException, PathName


class ConvertError(BatCaveException):
    """File Conversion Exceptions.

    Attributes:
        BACKUP_EXISTS: The backup file already exists.
    """
    BACKUP_EXISTS = BatCaveError(1, Template('File already exists: $file'))


class PackError(BatCaveException):
    """File Packing Exceptions.

    Attributes:
        INVALID_TYPE: An invalid archive type was specified.
        NO_FILES: No files were found to pack into the archive.
    """
    INVALID_TYPE = BatCaveError(1, Template('Unknown archive type: $arc_type'))
    NO_FILES = BatCaveError(2, 'No files to add')


COMPRESSION_TYPE = {'gz': 'gz', 'tgz': 'gz',
                    'bz': 'bz2', 'tbz': 'bz2', 'bz2': 'bz2',
                    'xz': 'xz', 'txz': 'xz'}
ConversionMode = Enum('ConversionMode', ('to_unix', 'to_dos'))


class CompressedFile:  # pylint: disable=too-few-public-methods
    """Class to add support for compressed file types which are missing some methods."""

    def __init__(self, filename: PathName, /, **_unused_attr):
        """
        Args:
            filename: The name of the compressed file.
            **attr: The list of attributes to pass to the base class.

        Attributes:
            _filename: The value of the filename argument.
        """
        self._filename = Path(filename)

    def namelist(self) -> Tuple[str]:
        """Return the name of the file as the first item in a tuple.

        Returns:
            The name of the file.
        """
        return (self._filename.stem,)


class BatCaveGzipFile(GzipFile, CompressedFile):  # pylint: disable=too-many-ancestors
    """Add CompressedFile class methods to the GzipFile class."""


class BatCaveBZ2File(BZ2File, CompressedFile):  # type: ignore[misc]  # pylint: disable=too-many-ancestors
    """Add CompressedFile class methods to the BZ2File class."""


PACKER_CLASSES = {'zip': ZipFile, 'gz': BatCaveGzipFile, 'bz2': BatCaveBZ2File, 'xz': LZMAFile}


def eol_convert(filename: PathName, mode: ConversionMode, /, *, backup: bool = True) -> None:
    """Perform end-of-line conversions from Windows to UNIX or vice versa.

    Attributes:
        filename: The file to convert.
        mode: The direction of the conversion. Must be a member of ConversionMode.
        backup (optional, default=True): If True, creates a backup of filename as filename.bak.

    Returns:
        Nothing.

    Raises:
        ConvertError.BACKUP_EXISTS: If backup is True and the backup file already exists.
    """
    filename = Path(filename)
    if backup:
        if (backup_file := filename.parent / f'{filename.name}.bak').exists():
            raise ConvertError(ConvertError.BACKUP_EXISTS, file=backup_file)
        copy(filename, backup_file)

    with open(filename, 'rb') as stream:
        data = stream.read().replace(b'\r\n', b'\n')

    if mode == ConversionMode.to_dos:
        data = data.replace(b'\n', b'\r\n')
    with open(filename, 'wb') as stream:
        stream.write(data)


def pack(archive_file: PathName, items: Iterable, /, item_location: Optional[PathName] = None, *,  # pylint: disable=too-many-locals,too-many-branches
         archive_type: str = '', ignore_empty: bool = True) -> None:
    """Create a compressed archive.

    Attributes:
        archive_file: The name of the archive file to create.
        items: The items to put in the archive.
        item_location (optional, default=None): The root location of the items. If None, the current directory
        archive_type (optional, default=None): the type of archive to create.
            If None, the type is derived from the archive_file extension.
        ignore_empty (optional, default=True): If True, allows creating an empty archive.

    Returns:
        Nothing.

    Raises:
        PackError.NO_FILES: If ignore_empty is False and there are no files to place in the archive.
    """
    archive = Path(archive_file)
    archive_type = archive.suffix.lstrip('.') if not archive_type else archive_type

    if item_location:
        pushd(item_location)

    tar_name = Path()
    tar_bug = False
    pkg_file = None
    if archive_type == 'zip':
        pkg_file = PACKER_CLASSES[archive_type](archive, 'w', ZIP_DEFLATED)
        adder = 'write'
    else:
        if compression := COMPRESSION_TYPE.get(archive_type, ''):
            tar_name = archive.with_suffix('.tar')
            tar_bug = True
        else:
            tar_name = archive

        pkg_file = tar_open(tar_name, mode='w:' + compression)  # type: ignore[call-overload]  # pylint: disable=consider-using-with
        adder = 'add'

    added = False
    for glob_item in [Path(i) for i in items]:
        for item in glob_item.parent.glob(glob_item.name):
            added = True
            if (adder == 'write') and item.is_dir():
                for (root, _unused_dirs, files) in item.walk():
                    for files_name in files:
                        getattr(pkg_file, adder)(Path(root, files_name))
            else:
                getattr(pkg_file, adder)(item)
    pkg_file.close()
    if (not ignore_empty) and (not added):
        raise PackError(PackError.NO_FILES)

    if tar_bug:
        if archive.exists():
            archive.unlink()
        tar_name.rename(archive)

    if item_location:
        popd()


def prune(directory: PathName, age: int, exts: Optional[Iterable[str]] = None,
          force: bool = False, ignore_case: bool = False, log_handle: Optional[str] = None) -> None:
    """Prune a directory of files or directories based on age or count.

    Args:
        age: The number days after which to prune files.
        directory: The directory from which to prune files.
        exts (optional, default=all): The extensions to prune.
        force (optional, default=False): If true, ignore permissions restricting removal.
        ignore_case (optional, default=False): If true, ignore case in extensions.
        log_handle (optional, default=None): If not None, status will be logged to the specified log handle.

    Returns:
        Nothing.
    """
    logger = getLogger(log_handle) if log_handle else None
    age_from = time()
    ext_list = [ext.lower() for ext in exts] if (exts and ignore_case) else exts
    target = Path(directory)
    if logger:
        logger.info('Removing %s in %s older than %d days',
                    ('/'.join(exts) + ' files') if exts else 'all', directory, age)
    if force:
        current_mode = target.stat().st_mode
        target.chmod(S_IRWXU)
    for item in target.iterdir():
        item_ext = item.suffix
        if ignore_case:
            item_ext = item_ext.lower()
        if (ext_list and (item_ext not in ext_list)) or not item.is_file():
            continue

        item_age = abs(int((age_from - item.stat().st_mtime) / 86400))
        if item_age > age:
            if logger:
                logger.info('  removing %s...', item.name)
            if force:
                item.chmod(S_IRWXU)
            item.unlink()
    if force:
        target.chmod(current_mode)


def slurp(filename: PathName, /) -> List[str]:
    """Return all the lines of a file as a list.

    Args:
        filename: The filename to return the lines from.

    Returns:
        The list of lines from the file.
    """
    return list(open(filename, encoding=DEFAULT_ENCODING))


def spew(filename: PathName, outlines: Iterable, /) -> None:
    """Write the list of lines to a file.

    Args:
        filename: The filename to which to write the lines.
        outlines: The lines to write.

    Returns:
        Nothing.
    """
    with open(filename, 'w', encoding=DEFAULT_ENCODING) as output_stream:
        output_stream.writelines(outlines)


def unpack(archive_file: PathName, dest: Optional[PathName] = None, /, *, archive_type: str = '') -> None:
    """Extract the contents of a compressed file.

    Attributes:
        archive_file: The name of the archive file
        dest (optional, default=None): The root location to which to extract. If None, the current directory
        archive_type (optional, default=None): the type of archive being extracted.
            If None, the type is derived from the archive_file extension.

    Returns:
        Nothing.

    Raises:
        PackError.INVALID_TYPE: If the archive_type is unknown.
    """
    archive = Path(archive_file).resolve()
    if not archive_type:
        archive_type = 'tar' if ('.tar.' in archive.name) else archive.suffix.lstrip('.')

    if dest:
        use_dest = Path(dest)
        use_dest.mkdir(parents=True, exist_ok=True)
        pushd(use_dest)

    lister = extractor = ''
    pkg_file: Any = None
    match archive_type:
        case 'bz2' | 'gz' | 'xz' | 'zip':
            pkg_file = PACKER_CLASSES[archive_type](archive)
            lister = 'namelist'
            extractor = 'read'
        case 'tar':
            pkg_file = tar_open(archive)  # pylint: disable=consider-using-with
            lister = 'getmembers'
            extractor = 'extract'
        case _:
            raise PackError(PackError.INVALID_TYPE, arc_type=archive_type)

    for member_info in getattr(pkg_file, lister)():
        member_path = Path(member_info.name if (archive_type == 'tar') else member_info)
        data = getattr(pkg_file, extractor)(member_info)
        if extractor == 'read':
            member_path.parent.mkdir(parents=True, exist_ok=True)
            if not (member_path.is_dir() or member_info.endswith('/')):
                with open(member_path, 'wb') as member_file:
                    member_file.write(data)
    pkg_file.close()

    if dest:
        popd()

# cSpell:ignore topdown



"""This module provides utilities for managing file expansions.

An instantiation of this class will convert an XML into the procedure object::

    <procedure schema = "{schema-version}" >
        <header>HEADER</header>

        <flags>
            <FlagName question="Some Question">True|False, Yes|No,  1,0</FlagName>
        </flags>

        <environments>
            <common>
                <VariableName1>Variable Value 1</VariableName1>
            </common>
            <Environment1>
                <VariableName2>Variable Value 2</VariableName1>
            <Environment2
        </environments>

        <directories>
            <directory>Directory - Path</directory>
        </directories>

        <steps>
            <step condition="Optional Condition Expression" vars="var1=val1, var2=val2", repeat="var=val1, val2">
                "Multi-Step Text"
                <step>Step Text</step>
                <step import="Library Step Name">Alternate Step Text</step>
            </step>
        </steps>

        <step-library>
            <step name="Import Name" />
        </step-library>
    </procedure>
"""

# Import standard modules
from collections import OrderedDict as odict
from copy import deepcopy
from enum import Enum
from pathlib import Path
from re import compile as re_compile
from shutil import copyfile
from string import Template
from typing import cast, Any, Dict, List, Match, Optional, Sequence, Tuple
from xml.etree.ElementTree import fromstringlist as xml_parse, Element

# Import internal modules
from .fileutil import slurp
from .lang import DEFAULT_ENCODING, is_debug, str_to_pythonval, BatCaveError, BatCaveException, PathName, WIN32

OutputFormat = Enum('OutputFormat', ('text', 'html', 'csv'))


class ExpanderError(BatCaveException):
    """Expansion Exceptions.

    Attributes:
        NO_POST_DELIMITER: No postfix delimiter was found.
        NO_REPLACEMENT: No replacement was found for the specified variable
    """
    NO_POST_DELIMITER = BatCaveError(1, Template('No closing delimiter found in $substr'))
    NO_REPLACEMENT = BatCaveError(2, Template('No replacement found for variable ($var) in: $thing'))


class ProcedureError(BatCaveException):
    """Procedure Exceptions.

    Attributes:
        BAD_ENVIRONMENT: The requested environment is not valid.
        BAD_FLAG: The specified flag value is not value.
        BAD_FORMAT: The requested output format is not valid.
        BAD_LIBRARY: The requested library is not valid.
        BAD_SCHEMA: The procedure schema is not supported.
        EXPANSION_ERROR: The was an unspecified error during expansion.
    """
    BAD_ENVIRONMENT = BatCaveError(1, Template('Unknown environment requested: $env.'))
    BAD_FLAG = BatCaveError(2, Template('Invalid value for flag: $value'))
    BAD_FORMAT = BatCaveError(3, Template('Unknown output format requested: $format'))
    BAD_LIBRARY = BatCaveError(4, Template('Unable to locate import: $lib.'))
    BAD_SCHEMA = BatCaveError(5, Template('Procedure specified in wrong schema ($schema). Please use schema $expected.'))
    EXPANSION_ERROR = BatCaveError(6, Template('$err\n  On line: $text'))


class Formatter:
    """Class to hold formatting information.

    Attributes:
        _LINK_PRELIM: The prefix to indicate a hyperlink during formatting.
    """
    _LINK_PRELIM = '{link:'

    def __init__(self, output_format: OutputFormat, /):
        """
        Args:
            output_format: The output format.

        Attributes:
            format: The value of the name argument.
            count: This current character count on the line.
            keeper: The indentation stack
            level: The current indentation level.
            link_regex: The regular expression to locate a hyperlink.
            prefix: The line prefix for the current line.

        Raises:
            ProcedureError.BAD_FORMAT: If the requested output format is invalid.
        """
        self.format = output_format
        self.level = 0
        self.count = 1
        self.keeper: List[Tuple[int, str]] = []
        self.prefix = ''
        self.link_regex = re_compile(f'\\{self._LINK_PRELIM}(.+?)(\\|(.+))?\\}}')

        match self.format:
            case OutputFormat.csv | OutputFormat.text:
                self.bos = self.eos = self._bol = ''
                self.eol = '\n'
            case OutputFormat.html:
                self.bos = '<ul>'
                self.eos = '</ul>'
                self._bol = '<h2><li>'
                self.eol = '</li></h2>'
            case _:
                raise ProcedureError(ProcedureError.BAD_FORMAT, format=self.format)

    @property
    def bol(self) -> str:
        """A read-only property which returns the beginning of line formatting."""
        if self.level == 0:
            sep = ',' if (self.format == OutputFormat.csv) else '. '
            match self.format:
                case OutputFormat.csv | OutputFormat.text:
                    return chr(64 + self.count) + sep
                case OutputFormat.html:
                    return self._bol
                case _:
                    raise ProcedureError(ProcedureError.BAD_FORMAT, format=self.format)
        else:
            if self.format == OutputFormat.csv:
                space = ''
                sep = ','
            else:
                space = '    ' * self.level
                sep = ': '
            return f'{self._bol}{space}{self.prefix}{self.count}{sep}'
        raise ProcedureError(ProcedureError.BAD_FORMAT, format=self.format)

    def format_hyperlinks(self, line: str, /) -> str:
        """Format the hyperlinks in a line.

        Args:
            line: The line for which to format hyperlinks.

        Returns:
            The formatted line.

        Raises:
            ProcedureError.BAD_FORMAT: If the requested format is not valid.
        """
        if not line or (self._LINK_PRELIM not in line):
            return line

        if not (match := self.link_regex.search(line)):
            raise ProcedureError(ProcedureError.BAD_FORMAT, format=self.format)
        replace_what = match.group(0)
        link = match.group(1)
        text = match.group(3) if (len(match.groups()) == 3) else ''
        replace_with = ''
        match self.format:
            case OutputFormat.csv:
                replace_with = f'"=HYPERLINK(""{link}"", ""{text}"")"' if text else link
            case OutputFormat.text:
                replace_with = f'{text} ({link})' if text else link
            case OutputFormat.html:
                text = text if text else link
                replace_with = f'<a href="{link}">{text}</a>'
            case _:
                raise ProcedureError(ProcedureError.BAD_FORMAT, format=self.format)
        return line.replace(replace_what, replace_with)

    def increment(self) -> None:
        """Increment the counter at the current indentation level.

        Returns:
            Nothing.
        """
        self.count += 1

    def indent(self) -> None:
        """Increment the indentation level.

        Returns:
            Nothing.
        """
        self.keeper.append((self.count, self.prefix))
        if self.level > 0:
            self.prefix = f'{self.keeper[-1][1]}{self.count}.'
        elif self.format == OutputFormat.html:
            self._bol = '<li style="color:white"><span style="color:black">'
            self.eol = '</span></li>'
        self.level += 1
        self.count = 1

    def outdent(self) -> None:
        """Decrement the indentation level.

        Returns:
            Nothing.
        """
        self.level -= 1
        (self.count, self.prefix) = self.keeper.pop()
        if self.level > 0:
            self.count += 1
        elif self.format == OutputFormat.html:
            self._bol = '<h2><li>'
            self.eol = '</li></h2>'


class Expander:
    """Class to handle interpolation of strings and files.

    Attributes:
        _PRELIM_DEFAULT: The prefix for a variable to be expanded.
        _POSTLIM_DEFAULT: The suffix for a variable to be expanded.
    """
    _PRELIM_DEFAULT = '{var:'
    _POSTLIM_DEFAULT = '}'

    def __init__(self, *, var_dict: Optional[Dict[str, str]] = None, var_props: Any = None, prelim: str = _PRELIM_DEFAULT, postlim: str = _POSTLIM_DEFAULT):
        """
        Args:
            var_dict: A dictionary of expansion variables.
            var_prop: An object with properties that resolve variables.
            prelim: The leading string to identify an expansion.
            postlim: The training string to identify an expansion.

        Attributes:
            postlim: The value of the postlim argument.
            prelim: The value of the prelim argument.
            re_var: The regular expression to identify a variable to expand.
            var_dict: The value of the var_dict argument.
            var_prop: The value of the var_prop argument.
        """
        self.var_dict = var_dict if var_dict else {}
        self.var_props = var_props if (isinstance(var_props, (list, tuple))) else [var_props]
        self.prelim = prelim
        self.postlim = postlim
        prelim_re = self.prelim
        postlim_re = self.postlim
        for spec in '.^$*+?|!/{}[]()<>:':
            prelim_re = prelim_re.replace(spec, '\\' + spec)
            postlim_re = postlim_re.replace(spec, '\\' + spec)
        self.re_var = re_compile(f'{prelim_re}([.a-zA-Z0-9_:]+){postlim_re}')

    def evaluate_expression(self, expression: str, /) -> bool:
        """Evaluate an expression in the expansion.

        Args:
            expression: The expression to evaluate.

        Returns:
            The evaluated expression.

        Raises:
            ExpanderError.NO_REPLACEMENT: If no replacement was found for the requested expansion variable.
        """
        if is_debug('EXPANDER'):
            print(f'Expanding (raw expression) "{expression}"')

        if isinstance(expression, bool):
            return expression
        if not expression:
            return True

        expression = self.expand(expression)
        synonyms = {' not ': ('!', '-not'),
                    ' and ': ('&&', ',', '-and'),
                    ' or ': ('||', '-or')}
        for (operator, syn_list) in synonyms.items():
            for synonym in syn_list:
                expression = expression.replace(synonym, operator)
        try:
            if is_debug('EXPANDER'):
                print(f'Expanding (corrected expression) "{expression}"')
            result = eval(expression, self.var_dict)  # pylint: disable=eval-used
            if is_debug('EXPANDER'):
                print(f'Expanding (evaluated expression) "{result}"')
            if isinstance(result, str):
                result = str_to_pythonval(result)
            if is_debug('EXPANDER'):
                print(f'Returning: "{result}"')
        except NameError as err:
            raise ExpanderError(ExpanderError.NO_REPLACEMENT, var=str(err), thing=expression) from err
        return result

    def expand(self, thing: Any, /) -> Any:
        """Perform an expansion on a Python object.

        Args:
            thing: The object on which to perform an expansion.

        Returns:
            The object with expansions performed.

        Raises:
            ExpanderError.NO_POST_DELIMITER: If the closing delimiter was not found.
            ExpanderError.NO_REPLACEMENT: If no replacement was found for the requested expansion variable.
        """
        if isinstance(thing, tuple):
            return {self.expand(i) for i in thing}

        if isinstance(thing, list):
            return [self.expand(i) for i in thing]

        if isinstance(thing, dict):
            return {self.expand(k): self.expand(v) for (k, v) in thing.items()}

        if (thing is None) or isinstance(thing, (bool, int, float)):
            return thing

        var = substr = replacer = ''
        while self.prelim in thing:
            try:
                var = cast(Match[str], self.re_var.search(thing)).group(1)
            except AttributeError as err:
                prelim_index = thing.index(self.prelim)
                substr = thing[prelim_index:prelim_index + 200]
                raise ExpanderError(ExpanderError.NO_POST_DELIMITER, substr=substr) from err

            fail = True
            if var in self.var_dict:
                replacer = self.var_dict[var]
                fail = False
            else:
                for var_props in self.var_props:
                    if hasattr(var_props, var):
                        replacer = getattr(var_props, var)
                        fail = False
                        break

            if fail:
                raise ExpanderError(ExpanderError.NO_REPLACEMENT, var=var, thing=thing)

            thing = thing.replace(f'{self.prelim}{var}{self.postlim}', str(replacer))
        return thing

    def expand_directory(self, source_dir: PathName, /, target_dir: Optional[PathName] = None, *,
                         ignore_files: Sequence[str] = tuple(), no_expand_files: Sequence[str] = tuple(), err_if_exists: bool = True) -> None:
        """Perform an expansion on an entire directory tree.

        Args:
            source_dir: The name of the directory on which to perform the expansions.
            target_dir (optional, default=None): The name of the output directory for the expansion results if not None,
                otherwise the current directory is used.
            ignore_files (optional, default=None): A list of files for which should be ignored and will not be in the output directory.
            no_expand_files (optional, default=None): A list of files which expansion should not be performed but which will be copied to the output directory.
            err_if_exists (optional, default=True): If True, raise an error if the output directory exists.

        Returns:
            Nothing.
        """
        source_dir = Path(source_dir).resolve()
        target_dir = Path(target_dir).resolve() if target_dir else Path.cwd()
        if is_debug('EXPANDER'):
            print('Using source directory:', source_dir)
            print('Using target directory:', target_dir)
        target_dir.mkdir(parents=True, exist_ok=not err_if_exists)
        for (root, _unused_dirs, files) in source_dir.walk():
            for file_name in files:
                if is_debug('EXPANDER'):
                    print('Checking', file_name, 'against', ignore_files)
                if file_name in ignore_files:
                    continue
                source_file = Path(root, file_name)
                target_file = target_dir / str(root).replace(str(source_dir), '').lstrip('\\' if WIN32 else '/') / file_name
                if file_name in no_expand_files:
                    if is_debug('EXPANDER'):
                        print(f'Copying {source_file} to {target_file} (root={root})')
                    copyfile(source_file, target_file)
                else:
                    if is_debug('EXPANDER'):
                        print(f'Expanding {source_file} to {target_file} ({root=})')
                    self.expand_file(source_file, target_file)

    def expand_file(self, in_file: PathName, out_file: PathName, /) -> None:
        """Perform an expansion on an entire file.

        Args:
            in_file: The name of the file on which to perform the expansions.
            out_file: The name of the output file for the expansion results.

        Returns:
            Nothing.
        """
        with open(in_file, encoding=DEFAULT_ENCODING) as in_stream:
            Path(out_file).parent.mkdir(parents=True, exist_ok=True)
            with open(out_file, 'w', encoding=DEFAULT_ENCODING) as out_stream:
                for line in in_stream:
                    line = self.expand(line)
                    out_stream.write(line)


def file_expander(in_file: PathName, out_file: PathName, /, *, var_dict: Optional[Dict[str, str]] = None, var_props: Any = None) -> None:
    """Quick function for one-time file expansion.

    Args:
        in_file: The input file.
        out_file: The output file.
        var_dict (optional, default=None): If not None, provides a dictionary of expansion values.
        var_props (optional, default=None): If not None, provides an object with properties to be used as expansion values.

    Returns:
        Nothing.
    """
    Expander(var_dict=var_dict, var_props=var_props).expand_file(in_file, out_file)


class Procedure:
    """Class to create a universal abstract interface for a procedure.

    Attributes:
        _SCHEMA_ATTR: The XML tag which contains the schema version.
        _REQUIRED_PROCEDURE_SCHEMA: The schema supported by this class.

        _HEADER_TAG: The XML tag which indicates the procedure header.
        _FLAGS_TAG: The XML tag which indicates the procedure flags.
        _DIRECTORIES_TAG: The XML tag which indicates the procedure directories.
        _ENVIRONMENTS_TAG: The XML tag which indicates the procedure environments.
        _STEPS_TAG: The XML tag which indicates the procedure steps.
        _LIBRARY_TAG: The XML tag which indicates the procedure step library.

        _COMMON_ENVIRONMENT: The XML tag which indicates the procedure common environment.
        _ENVIRONMENT_VARIABLE: The XML tag which indicates an environment variable.
    """
    _COMMON_ENVIRONMENT = 'common'
    _DIRECTORIES_TAG = 'directories'
    _ENVIRONMENTS_TAG = 'environments'
    _ENVIRONMENT_VARIABLE = 'Environment'
    _FLAGS_TAG = 'flags'
    _HEADER_TAG = 'header'
    _LIBRARY_TAG = 'step-library'
    _REQUIRED_PROCEDURE_SCHEMA = 1
    _SCHEMA_ATTR = 'schema'
    _STEPS_TAG = 'steps'

    def __init__(self, proc_file: PathName, /, *, output_format: OutputFormat = OutputFormat.html, variable_overrides: Optional[Dict[str, str]] = None):  # pylint: disable=too-many-locals
        """
        Args:
            proc_file: The procedure file.
            output_format: The output format.
            variable_overrides: A dictionary of values used to override values defined in the procedure file.

        Attributes:
            directories: The list of directories defined in the procedure file.
            environments: The dictionary of directories defined in the procedure file.
            expander: The expander object used to expand the procedure file.
            formatter: The formatter object used to format the output.
            header: The header of the output.
            library: The list of step libraries defined in the procedure file.
            output_format: The value of the output_format argument.
            steps: The list of steps defined in the procedure file.

        Raises:
            ProcedureError.BAD_SCHEMA: If the schema of the procedure file is not supported.
        """
        self.output_format = output_format
        self.formatter: Formatter
        self.expander: Expander

        if (schema := str_to_pythonval((xml_root := xml_parse(slurp(proc_file))).get(self._SCHEMA_ATTR, '0'))) != self._REQUIRED_PROCEDURE_SCHEMA:
            raise ProcedureError(ProcedureError.BAD_SCHEMA, schema=schema, expected=self._REQUIRED_PROCEDURE_SCHEMA)
        self.header = str(xml_root.findtext(self._HEADER_TAG)) if xml_root.findtext(self._HEADER_TAG) else ''
        flags = {f.tag: parse_flag(str(f.text)) for f in list(flags_element)} if (flags_element := xml_root.find(self._FLAGS_TAG)) else {}  # pylint: disable=used-before-assignment
        self.directories = [str(d.text) for d in list(directories_element)] if (directories_element := xml_root.find(self._DIRECTORIES_TAG)) else []  # pylint: disable=used-before-assignment
        self.steps = [Step(s) for s in list(steps_element)] if (steps_element := xml_root.find(self._STEPS_TAG)) else []  # pylint: disable=used-before-assignment
        self.library = {r.attrib[Step.NAME_ATTR]: Step(r) for r in list(library_element)} if (library_element := xml_root.find(self._LIBRARY_TAG)) else {}  # pylint: disable=used-before-assignment

        environments_element = xml_root.find(self._ENVIRONMENTS_TAG)
        self.environments: Dict = {e.tag: {v.tag: (v.text if v.text else '') for v in list(e)} for e in list(environments_element)} if environments_element else {}

        common_environment: Dict
        if self._COMMON_ENVIRONMENT in self.environments:
            common_environment = self.environments[self._COMMON_ENVIRONMENT]
            del self.environments[self._COMMON_ENVIRONMENT]
        else:
            common_environment = {}

        # Convert the flags to True/False variables in the common environment
        for (flag_name, flag_value) in flags.items():
            common_environment[flag_name] = flag_value

        # Create a default value of False for IsEnvironment of every environment
        for env in self.environments:
            common_environment['Is' + env] = False

        # Update the environments with the common environment values and
        # set the Environment variable and
        # set IsEnvironment to True for that environment
        for env in self.environments:
            env_dict = common_environment | self.environments[env]
            if variable_overrides:
                env_dict |= variable_overrides
            self.environments[env] = env_dict
            if self._ENVIRONMENT_VARIABLE not in self.environments[env]:
                self.environments[env][self._ENVIRONMENT_VARIABLE] = env
            self.environments[env]['Is' + env] = True

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        return False

    def dump(self) -> Dict[str, str | Sequence | Dict]:
        """Dump out the procedure contents.

        Returns:
            The contents of the procedure as an ordered dictionary.
        """
        result: Dict = odict()
        result[self._HEADER_TAG] = self.header
        result[self._DIRECTORIES_TAG] = self.directories
        result[self._ENVIRONMENTS_TAG] = {e: v for (e, v) in self.environments.items()}  # pylint: disable=unnecessary-comprehension
        result[self._LIBRARY_TAG] = {r: s.dump() for (r, s) in self.library.items()}
        result[self._STEPS_TAG] = [s.dump() for s in self.steps]
        return result

    def expand(self, text: str, /) -> str:
        """Expand the Procedure.

        Args:
            text: The text of the procedure.

        Returns:
            The expanded procedure.

        Raises:
            ProcedureError.EXPANSION_ERROR: If there is an missing value expansion error.
        """
        try:
            return self.expander.expand(text)
        except ExpanderError as err:
            if err.code == ExpanderError.NO_REPLACEMENT.code:
                raise ProcedureError(ProcedureError.EXPANSION_ERROR, err=str(err), text=text) from err
            raise

    def expand_directories(self, env: str, /, destination_root: PathName, *, source_root: PathName = Path(), err_if_exists: bool = True) -> None:
        """Perform variable expansion on the directories defined in the procedure.

        Args:
            env: The environment for which to expand the directories.
            destination_root: That destination directory for the expansion.
            source_root (optional, default=None): Defined for recursion.
            err_if_exists (optional, default=True): If True, raise an error if destination_root exists.

        Returns:
            Nothing.
        """
        self.setup_expander(env)
        for dirname in self.directories:
            if not (dirpath := Path(dirname)).is_absolute():
                dirpath = Path(source_root) / dirpath
            self.expander.expand_directory(dirpath, Path(destination_root, dirname), err_if_exists=err_if_exists)

    def format(self, text: str, /) -> str:
        """Format an output line including hyperlinks.

        Args:
            text: The line of text to format.

        Returns:
            The formatted output line.
        """
        return self.formatter.format_hyperlinks(self.expand(text))

    def realize(self, env: str, /) -> str:
        """Realize the procedure for the specified environments based on the variables.

        Args:
            env: The environment for which to realize the procedure.

        Returns:
            The realized procedure.

        Raises:
            ProcedureError.BAD_FORMAT: If the format type is not defined.
        """
        header = footer = ''
        self.formatter = Formatter(self.output_format)
        self.setup_expander(env)
        match self.output_format:
            case OutputFormat.csv:
                header = f',{self.header.strip()} for {{var:Environment}}\n'
                footer = ''
            case OutputFormat.text:
                header = f'{self.header.strip()} for {{var:Environment}}\n'
                footer = ''
            case OutputFormat.html:
                header = '<html><meta http-equiv="Content-Type" content="text/html;charset=utf-8"><body><center>'
                header += f'<h1>{self.header} for {{var:Environment}}</h1></center><ol type="A">'
                footer = '</ol></body></html>'
            case _:
                raise ProcedureError(ProcedureError.BAD_FORMAT, format=self.output_format)

        content = ''
        for step in self.steps:
            text = self.realize_step(step)
            if text:
                content += self.formatter.eol + text
                self.formatter.increment()
        return self.format(header) + content + footer

    def realize_step(self, step: 'Step', /) -> str:  # pylint: disable=too-many-branches
        """Realize a step in the procedure.

        Args:
            step: The step to realize.

        Returns:
            The realized step.

        Raises:
            ProcedureError.BAD_LIBRARY: If the step specifies a library that is not defined.
        """
        if not self.expander.evaluate_expression(step.condition):
            return ''

        expander_vars_keeper = None
        if step.vars:
            expander_vars_keeper = deepcopy(self.expander.var_dict)
            self.expander.var_dict |= step.vars

        if step.lib_import:
            if step.lib_import not in self.library:
                raise ProcedureError(ProcedureError.BAD_LIBRARY, lib=step.lib_import)
            lib_step = deepcopy(self.library[step.lib_import])
            lib_step.repeat = step.repeat
            lib_step.vars |= step.vars
            if step.text:
                lib_step.text = step.text
            elif not lib_step.text:
                lib_step.text = lib_step.name
            step = lib_step

        output = ''
        if step.repeat:
            (variable, values_as_str) = step.repeat.split('=')
            values = [v.strip() for v in self.expand(values_as_str).split(',')]
            step_copy = deepcopy(step)
            step_copy.repeat = ''
            for value in values:
                step_copy.vars[variable] = value
                output += self.realize_step(step_copy)
        elif step.sub_steps:
            start_of_step = self.formatter.bol + self.format(step.text) + self.formatter.eol + self.formatter.bos
            self.formatter.indent()
            for sub_step in step.sub_steps:
                output += self.realize_step(sub_step)
            if output:
                output = start_of_step + output + self.formatter.eos
            self.formatter.outdent()
        else:
            if output := self.format(step.text):
                output = self.formatter.bol + output + self.formatter.eol
                self.formatter.increment()

        if expander_vars_keeper:
            self.expander.var_dict = expander_vars_keeper
        return output

    def setup_expander(self, environment: str, /) -> None:
        """Setup the Expander for the requested environment.

        Args:
            environment: The environment for which to setup the expander.

        Returns:
            Nothing.

        Raises:
            ProcedureError.BAD_ENVIRONMENT: If the requested environment is not defined.
        """
        if environment not in self.environments:
            raise ProcedureError(ProcedureError.BAD_ENVIRONMENT, env=environment)
        self.expander = Expander(var_dict=self.environments[environment])


class Step:  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a procedure step.

    Attributes:
        NAME_ATTR: The XML attribute to indicate the step name.
        _CONDITION_ATTR: The XML attribute to indicate the step condition.
        _IMPORT_ATTR: The XML attribute to indicate the step import.
        _REPEAT_ATTR: The XML attribute to indicate the step repeat condition.
        _VARS_ATTR: The XML attribute to indicate the step variables.
    """
    _CONDITION_ATTR = 'condition'
    _IMPORT_ATTR = 'import'
    _REPEAT_ATTR = 'repeat'
    _VARS_ATTR = 'vars'

    NAME_ATTR = 'name'

    def __init__(self, step_def: Element, /):
        """
        Args:
            step_def: The dictionary which defined the step.

        Attributes:
            condition: The condition the determines if the step is emitted.
            lib_import: The library that is imported for the step.
            name: The step name.
            repeat: The repeat parameters for the step.
            sub_steps: The list of sub_steps for the step.
            text: The text emitted for the step.
            vars: The dictionary of variables defined for the step.
        """
        self.condition = step_def.get(self._CONDITION_ATTR, '')
        self.lib_import = step_def.get(self._IMPORT_ATTR, '')
        self.name = step_def.get(self.NAME_ATTR, '')
        self.repeat = step_def.get(self._REPEAT_ATTR, '')
        self.text = step_def.text.strip() if step_def.text else ''
        self.sub_steps = [Step(s) for s in list(step_def)]
        self.vars = {v.split('=')[0].strip(): v.split('=')[1].strip() for v in var.split(',')} if (var := step_def.get(self._VARS_ATTR, '')) else {}  # pylint: disable=used-before-assignment

    def dump(self) -> List:
        """Dump out the step contents.

        Returns:
            The contents of the step as an list.
        """
        return [f'{self.text}: import={self.lib_import}: condition={self.condition}: repeat={self.repeat}: vars={self.vars}'] + [str(s.dump()) for s in self.sub_steps]


def parse_flag(flag: str, /) -> Any:
    """Evaluate a parsing flag.

    Args:
        flag: The flag to evaluate.

    Return:
        The evaluated value for the flag.

    Raises:
        ProcedureError.BAD_FLAG: If the evaluated flag is not of type bool.
    """
    if not isinstance(value := str_to_pythonval(flag.lower().replace('0', 'False').replace('1', 'True').replace('yes', 'True').replace('no', 'False')), bool):
        raise ProcedureError(ProcedureError.BAD_FLAG, value=flag)
    return value

# cSpell:ignore odict fileutil pythonval postlim



"""This module provides Python language utilities.

Attributes:
    DEFAULT_ENCODING: The default encoding used for text file operations.
    BATCAVE_HOME: The home directory of the module.
    FROZEN (bool): Is this module running in a frozen application. Quick version of sys.frozen
    VALIDATE_PYTHON (bool, default=True): Whether this module should validate the minimum version of Python when loaded.
    WIN32 (bool): Is this module running on a Windows system. Quick version of (sys.platform == 'win32')
"""

# Import standard modules
from dataclasses import dataclass
from os import getenv
from pathlib import Path, PurePath
from string import Template
import sys
from sys import executable, platform, version_info, path as sys_path
from typing import Any, Dict, Iterable, List, Optional, Tuple

# Import third-party modules
from dotmap import DotMap
from yaml import safe_dump as yaml_dump, safe_load as yaml_load

# Useful constants
DEFAULT_ENCODING = 'UTF-8'
FROZEN = getattr(sys, 'frozen', False)
BATCAVE_HOME = Path(executable).parent if FROZEN else Path(sys_path[0])
VALIDATE_PYTHON = True
WIN32 = platform == 'win32'

type CommandResult = str | List[str]
type MessageString = str | Template
type PathName = str | Path | PurePath


class MsgStr:
    """Class to create a universal abstract interface for message strings.

    This class is only useful when subclassed where the subclass simply defines the _messages.

    Attributes:
        _message: A dictionary of messages provided by subclasses.

    Example:
        class MyMsg(MsgStr):
            _messages = {'Message1': 'This is just a string',
                            'Message2': Template('This is a $what template)}

        where messages are retrieved with
            MyMsg().Message1
            MyMsg(what='this').Message2
    """
    _messages: Dict[str, str | Template] = {}

    def __init__(self, instr: str | Template = '', transform: str = '', **variables):
        """
        Args:
            instr (optional, default=''): The input message string.
            transform (optional, default=None): A string method used to transform the input message string on output.
            variables (optional): A dictionary of variables to pass to the string.Template.substitute method.

        Attributes:
            _str: The value of the instr argument.
            _transform: The value of the transform argument.
            _vars: The value of the variables argument.
        """
        self._str = instr
        self._transform = transform
        self._vars = variables

    def __getattr__(self, attr: str) -> str:
        if attr in self._messages:
            return self._self_to_str(self._messages[attr])
        raise AttributeError(f"'{type(self)}' object has no attribute '{attr}'")

    def __str__(self):
        return self._self_to_str(self._str)

    def _self_to_str(self, _str: MessageString, /) -> str:
        """Convert this message to a string and apply any transforms.

        Returns:
            The string message.
        """
        return_str: str = _str.substitute(self._vars) if isinstance(_str, Template) else _str
        if self._transform:
            return_str = getattr(return_str, self._transform)()
        return return_str


class BatCaveException(Exception, MsgStr):
    """A base class to provide easier Exception management."""
    def __init__(self, err_obj: 'BatCaveError', /, **variables):
        """
        Args:
            err_obj: The input message string.
            variables (optional): A dictionary of variables to pass to the string.Template.substitute method.

        Attributes:
            vars: The value of the variables argument.
            _err_obj: The value of the err_obj argument.
        """
        Exception.__init__(self, err_obj, variables)
        MsgStr.__init__(self, err_obj.msg, **variables)
        self._err_obj = err_obj
        self.vars = variables

    def __str__(self):
        return MsgStr.__str__(self)

    code = property(lambda s: s._err_obj.code, doc='A read-only property which returns the error code from the error object.')


@dataclass(frozen=True)
class BatCaveError:
    """A class to provide an interface for inspecting exceptions.

        Attributes:
            code: A unique error code for this error.
            msg: A user-facing message for this error.
    """
    code: int
    msg: MessageString


class PythonVersionError(BatCaveException):
    """Invalid Python Version Exception.

    Attributes:
        BAD_VERSION: The version of Python is too low.
    """
    BAD_VERSION = BatCaveError(1, Template('Python $needed required but $used used'))


def bool_to_str(expr: bool | str, /) -> str:
    """Converts an expression to a lowercase boolean string value.

    Args:
        expr: The expression to convert.

    Returns:
        'true' if the expression evaluates to True, 'false' otherwise.
    """
    return str(bool(expr)).lower()


def dotmap_to_yaml(dotmap_thing: DotMap, yaml_file: PathName, /) -> None:
    """Convert a DotMap to a YAML file.

    Args:
        dotmap_thing: The DotMap to convert.
        yaml_file: The YAML file to which to write the DotMap.

    Returns:
        Nothing.
    """
    with open(yaml_file, 'w', encoding=DEFAULT_ENCODING) as yaml_stream:
        return yaml_dump(dotmap_thing.toDict(), yaml_stream)


def yaml_to_dotmap(yaml_info: str | PathName, /) -> DotMap:
    """Converts a YAML file to a DotMap.

    Args:
        yaml_info: The YAML to which to read into a DotMap. If this is a string it is presumed to be raw YAML
                    otherwise it is expected to be a Path object which can be open and read.

    Returns:
        A DotMap representing the YAML content.
    """
    if isinstance(yaml_info, str):
        return DotMap(yaml_load(yaml_info))
    with open(yaml_info, encoding=DEFAULT_ENCODING) as yaml_stream:
        return DotMap(yaml_load(yaml_stream))


def flatten(thing: Iterable[Iterable], /, *, recursive: bool = True) -> Iterable:
    """Flatten an iterable of iterables.

    Args:
        thing: The thing to be flattened.
        recursive (optional, default=True): Whether or not to recursively flatten the item.

    Returns:
        The final single depth item as the same type as thing.
    """
    flattened = False
    result = []
    for item in thing:
        try:
            result += [i for i in iter(item)]  # pylint: disable=unnecessary-comprehension
            flattened = True
        except TypeError:
            result.append(item)

    if recursive and flattened:
        return flatten(result)

    return type(thing)(result)  # type: ignore[call-arg]


def flatten_string_list(iter_of_string: Iterable[str], /, *, remove_newlines: bool = True) -> str:
    """Flatten an iterable of strings to a single string.

    Args:
        iter_of_string: The list of strings to be flattened to be flattened.
        remove_newlines (optional, default=True): Whether or not to remove newlines from the final list.

    Returns:
        The final string.
    """
    result = ''.join(flatten(iter_of_string))
    if remove_newlines:
        return result.replace('\n', '')
    return result


def is_debug(test_value: Optional[str] = None, /) -> bool:
    """Determine if the BATCAVE_DEBUG environment variable is set.

    Args:
        test_value (optional, default=False): If set, only return true if the value of test_value is in BATCAVE_DEBUG.

    Return:
        True if the OS environment variable BATCAVE_DEBUG is set, False otherwise.
    """
    return bool(debug_value := getenv('BATCAVE_DEBUG', '')) and ((not test_value) or (test_value in debug_value))


def str_to_pythonval(the_string: str, /, *, parse_python: bool = False) -> Any:
    """Converts a string to the closest Python object.

    Args:
        the_string: The string to evaluate.
        parse_python (optional, default=False): If the string contains a '~' character, try to convert it to a more complex python object.

    Returns:
        #. If the string represents an integer, return the value as an int.
        #. If the string represents an non-integer number, return the value as a float.
        #. If the string evaluates to 'None' (case-insensitive), return None.
        #. If the string evaluates to 'True' or 'False' (case-insensitive), return True/False.
        #. If parse_python is True and the_string contains '~':
            Split the_string on the first '~' and return the second part as the value of a type specified by the first part.

    Raises:
        ValueError: If the_string is not a string.
    """
    if not isinstance(the_string, str):
        raise ValueError

    if the_string.isdecimal():
        return int(the_string)

    if the_string.isdigit():
        return float(the_string)

    for (test, value) in (('none', None), ('true', True), ('false', False)):
        if the_string.lower() == test:
            return value

    if parse_python and '~' in the_string:
        (data_type, val) = the_string.split('~', 1)
        the_string = eval(f'{data_type}({val})')  # pylint: disable=eval-used

    if parse_python and '~' in the_string:
        (data_type, val) = the_string.split('~', 1)
        the_string = eval(f'{data_type}({val})')  # pylint: disable=eval-used

    return the_string


def validate_python(test_against: Tuple[int, int] = (3, 6), /) -> None:
    """Checks to make sure that a minimum version of Python is used.

    Args:
        test_against (optional, default=(3,7)): The value of Python to check.

    Returns:
        Nothing.

    Raises:
        PythonVersionError.BAD_VERSION: If the version is too low.
    """
    used = version_info[:2]
    needed = test_against if test_against else (3, 6)
    if used != needed:
        raise PythonVersionError(PythonVersionError.BAD_VERSION, used=used, needed=needed)


def xor(value1: Any, value2: Any, /) -> bool:
    """Perform a logical exclusive-or evaluation.

    Args:
        value1, value2: The values on which to perform the operation.

    Returns:
        The logical exclusive-or of the values.
    """
    return bool(value1) ^ bool(value2)

# cSpell:ignore batcave dotmap pythonval



"""This module provides a Pythonic interface to work with system utilities.

Attributes:
    PROG_FILES (dict): The default software installation location
        For 32-bit Windows systems this is the value of the ProgramFiles(x86) environment variable.
        For 64-bit Windows systems this is the value of the ProgramFiles environment variable.
        For all other systems it is /usr/local.
    S_660: A quick version of the UNIX 0660 mode.
    S_664: A quick version of the UNIX 0664 mode.
    S_770: A quick version of the UNIX 0770 mode.
    S_775: A quick version of the UNIX 0775 mode.
"""

# Import standard modules
import sys
from copy import copy as copy_object
from enum import Enum
from errno import EACCES, EAGAIN, ECHILD
from os import chdir, getenv, remove, unlink
from pathlib import Path
from shutil import rmtree, chown as os_chown
from stat import S_IRUSR, S_IWUSR, S_IRGRP, S_IWGRP, S_IROTH, S_IRWXU, S_IRWXG, S_IXOTH
from string import Template
from subprocess import Popen, PIPE
from typing import cast, Any, Dict, Callable, IO, Iterable, List, Optional, Tuple, TextIO

# Import internal modules
from .lang import DEFAULT_ENCODING, flatten_string_list, is_debug, BatCaveError, BatCaveException, CommandResult, PathName, WIN32

if sys.platform == 'win32':
    from msvcrt import locking, LK_NBLCK, LK_UNLCK  # pylint: disable=import-error
    PROG_FILES = {'32': Path(getenv('ProgramFiles(x86)', '')), '64': Path(getenv('ProgramFiles', ''))}
    geteuid = getgrnam = getpwnam = None  # Fix Pylance and mypy linting errors  # pylint: disable=invalid-name
else:
    from fcntl import lockf, LOCK_EX, LOCK_NB, LOCK_UN  # pylint: disable=import-error
    from grp import getgrnam  # pylint: disable=import-error
    from os import geteuid  # pylint: disable=no-name-in-module,ungrouped-imports
    from pwd import getpwnam  # pylint: disable=import-error
    PROG_FILES = {'32': Path('/usr/local')}
    PROG_FILES['64'] = PROG_FILES['32']

LockMode = Enum('LockMode', ('lock', 'unlock'))
S_660 = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP
S_664 = S_660 | S_IROTH
S_770 = S_IRWXU | S_IRWXG
S_775 = S_770 | S_IROTH | S_IXOTH

type ServerAuthType = str | Path | Tuple[str, str] | None


class CMDError(BatCaveException):
    """System Command Exceptions.

    Attributes:
        CMD_ERROR: Generic command error.
        CMD_NOT_FOUND: The command was not found.
        INVALID_OPERATION: The requested operation is not supported in the current context.
    """
    CMD_ERROR = BatCaveError(1, '')
    CMD_NOT_FOUND = BatCaveError(2, Template('Command not found when running: $cmd'))
    INVALID_OPERATION = BatCaveError(3, Template('$func is not supported for $context'))

    def __str__(self):
        if self._err_obj.code == CMDError.CMD_ERROR.code:
            err_lines = self.vars['err_lines'] if self.vars['err_lines'] else self.vars['outlines']
            return f'Error {self.vars['returncode']} when running: {self.vars['cmd']}\nError output:\n' + ''.join(err_lines)
        return BatCaveException.__str__(self)


class LockError(BatCaveException):
    """Lock File Exceptions.

    Attributes:
        NO_LOCK: There was a failure attempting to get a lock on the lock file.
    """
    NO_LOCK = BatCaveError(1, Template('Unable to get lock'))


class OSUtilError(BatCaveException):
    """Operating System Exceptions.

    Attributes:
        GROUP_EXISTS: The specified group already exists.
        INVALID_OPERATION: The requested operation is not supported on the current platform.
        USER_EXISTS: The specified user already exists.
    """
    GROUP_EXISTS = BatCaveError(1, Template('The group already exists: $group'))
    INVALID_OPERATION = BatCaveError(2, Template('Platform unsupported: $platform'))
    USER_EXISTS = BatCaveError(3, Template('The user already exists: $user'))


class LockFile:
    """Class to create a universal abstract interface for an OS lock file."""

    def __init__(self, filename: PathName, /, handle: Optional[TextIO] = None, *, cleanup: bool = True):
        """
        Args:
            filename: The file name for the lock file.
            handle (optional, default=None): The value of the file handle if the file is already open, otherwise the file specified in the filename will be opened.
            cleanup (optional, default=True): If True, the lock file will be removed when the lock is released.

        Attributes:
            _cleanup: The value of the cleanup argument.
            _fd: The fileno for the _fh attribute.
            _fh: The value of the handle argument if not None, otherwise the value of the handle for the opened filename argument.
            _filename: The value of the filename argument.
            _lock: The value to pass to the _locker method to lock the file.
            _locker: The method used to lock the file.
            _unlock: The value to pass to the _locker method to unlock the file.
        """
        self._filename = Path(filename)
        self._cleanup = cleanup
        self._fh = handle if handle else open(filename, 'w', encoding=DEFAULT_ENCODING)  # pylint: disable=consider-using-with
        self._fd = self._fh.fileno()
        if sys.platform == 'win32':
            self._locker = locking
            self._lock = LK_NBLCK
            self._unlock = LK_UNLCK
        else:
            self._locker = lockf
            self._lock = LOCK_EX | LOCK_NB
            self._unlock = LOCK_UN
        self.action(LockMode.lock)

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()
        return False

    def action(self, mode: LockMode, /) -> None:
        """Perform the requested action on the lock file.

        Args:
            mode: The action to perform on the lock file.

        Returns:
            Nothing.

        Raises:
            LockError.NO_LOCK: If it was not possible to obtain a system level lock on the lock file.
        """
        lock_mode = self._lock if (mode == LockMode.lock) else self._unlock
        try:
            self._locker(self._fd, lock_mode, 1)
        except IOError as err:
            if err.errno not in (EACCES, EAGAIN):
                raise
            raise LockError(LockError.NO_LOCK) from err

    def close(self) -> None:
        """Close the lock file.

        Returns:
            Nothing.
        """
        self.action(LockMode.unlock)
        self._fh.close()
        if self._cleanup:
            self._filename.unlink()


class SysCmdRunner:  # pylint: disable=too-few-public-methods
    """This class provides a simplified interface to sysutil.syscmd()."""

    def __init__(self, command: str, /, *args, show_cmd: bool = True, show_stdout: bool = True, syscmd_args: Optional[Dict[Any, Any]] = None, **kwargs):
        """
        Args:
            command: The command to run.
            show_cmd (optional, default=True): The default value of the show_cmd value passed to syscmd.
            show_stdout (optional, default=True): The default value of the show_stdout value passed to syscmd.
            syscmd_args (optional, default={}): The dictionary of other arguments passed to syscmd.
            *args (optional): A list of default arguments to use each time the command is run.
            **kwargs (optional): A list of default keys to use each time the command is run.

        Attributes:
            _command: The value of the command argument.
            _default_args: The value of the args argument.
            _default_kwargs: The value of the kwargs argument.
            _default_syscmd_kwargs: The value of the syscmd_args argument.
        """
        self._command = command
        self._default_args = list(args)
        self._default_kwargs = kwargs
        self._default_syscmd_args: Dict[Any, Any] = {'show_cmd': show_cmd, 'show_stdout': show_stdout}
        if syscmd_args:
            self._default_syscmd_args |= syscmd_args

    def run(self, *args, post_option_args: Optional[Dict] = None, syscmd_args: Optional[Dict[Any, Any]] = None, **kwargs) -> CommandResult:
        """Run the defined command with the additional specified arguments.

        Args:
            post_option_args (optional, default=[]): The list of args to pass after the options.
            syscmd_args (optional, default={}): The list of default args passed to syscmd.
            *args (optional, default=[]): Any extra arguments to pass to the command.
            **kwargs (optional, default={}): Any extra keyword arguments to pass to the command.

        Returns:
            The result of the syscmd call.
        """
        command_args = self._default_args + list(args)
        option_args = copy_object(self._default_kwargs) | kwargs
        for (arg, value) in option_args.items():
            arg_name = arg.replace('_', '-')
            if value is True:
                command_args.append(f'--{arg_name}')
            else:
                command_args.append(f'--{arg_name}={value}')
        if post_option_args:
            command_args += post_option_args
        all_syscmd_args = copy_object(self._default_syscmd_args)
        if syscmd_args:
            all_syscmd_args |= syscmd_args
        return syscmd(self._command, *command_args, **all_syscmd_args)


def chmod(dirname: PathName, mode: int, *, recursive: bool = False, files_only: bool = False) -> None:
    """Perform chmod recursively if requested.

    Args:
        dirname: The directory for which to set the mode.
        mode: The mode to set as would be specified for os.chmod().
        recursive (optional, default=False): Also perform setting recursively on all children.
        files_only (optional, default=False): Only affect files.

    Returns:
        Nothing.
    """
    dirname = Path(dirname)
    if not files_only:
        dirname.chmod(mode)
    if recursive:
        for (root, dirs, files) in dirname.walk():
            for pathname in (files if files_only else (dirs + files)):  # pylint: disable=superfluous-parens
                Path(root, pathname).chmod(mode)


def chown(pathname: PathName, user: str | int, group: Optional[str | int] = None, *, recursive: bool = False) -> None:
    """Perform chown and chgrp together, recursively if requested.

    Args:
        user (optional, default=None): The user to set as the owner, if specified.
        group (optional, default=None): The group to set as the owner, if specified.
        recursive (optional, default=False): Also perform the user/owner settings recursively on all children.

    Returns:
        Nothing.
    """
    os_chown(path := Path(pathname), user, group)
    if recursive:
        for (root, dirs, files) in path.walk():
            for sub_path in dirs + files:
                os_chown(Path(root, sub_path), user, group)


def create_group(group_name: str, /, *, exists_ok: bool = True) -> None:
    """Create the system group at the OS level.

    Args:
        group_name: The group to create.
        exists_ok (optional, default=True): Do not raise an error if the group exists.

    Returns:
        Nothing.

    Raises:
        OSUtilError.GROUP_EXISTS: If the group exists and exists_ok is False.
        OSUtilError.INVALID_OPERATION: If this is a Windows platform.

    Todo:
        Implement for the Windows platform.
    """
    if sys.platform == 'win32':
        raise OSUtilError(OSUtilError.INVALID_OPERATION, platform='Windows')
    try:
        getgrnam(group_name)
        if not exists_ok:
            raise OSUtilError(OSUtilError.GROUP_EXISTS, group=group_name)
    except KeyError:
        syscmd('groupadd', group_name)


def create_user(username: str, /, groups: Tuple = tuple(), *, exists_ok: bool = True) -> None:
    """Create the user account at the OS level.

    Args:
        username: The user account to create.
        exists_ok (optional, default=True): Do not raise an error if the user exists.

    Returns:
        Nothing.

    Raises:
        OSUtilError.USER_EXISTS: If the user exists and exists_ok is False.
        OSUtilError.INVALID_OPERATION: If this is a Windows platform.

    Todo:
        Implement for the Windows platform.
    """
    if sys.platform == 'win32':
        raise OSUtilError(OSUtilError.INVALID_OPERATION, platform='Windows')
    create_group(username, exists_ok=exists_ok)
    try:
        getpwnam(username)
        if not exists_ok:
            raise OSUtilError(OSUtilError.USER_EXISTS, group=username)
    except KeyError:
        groups_args = (('-G',) + groups) if groups else tuple()
        syscmd('useradd', username, '-g', username, *groups_args)


def is_user_administrator() -> bool:
    """Determines if the current user is an OS administrator.

    Args:
        None

    Returns:
        True if the user is an administrator, False otherwise.
    """
    if sys.platform == 'win32':
        try:
            syscmd('net', 'file')
        except CMDError as err:
            if 'Access is denied' in str(err):
                return False
            raise
    elif geteuid() != 0:
        return False
    return True


def rmpath(path_name: PathName, /) -> None:
    """Remove the specified path object. If a directory, remove recursively.

    Args:
        path_name: The name of the path object to remove.

    Returns:
        Nothing.
    """
    if (path_name := Path(path_name)).is_dir():
        rmtree_hard(path_name)
    else:
        path_name.unlink()


def rmtree_hard(tree: PathName, /) -> None:
    """Recursively, remove a directory and try to avoid non-fatal errors.

    Args:
        tree: The directory tree to remove.

    Returns:
        Nothing.
    """
    Path(tree).chmod(S_IRWXU)
    rmtree(tree, onexc=_rmtree_onerror)


def _rmtree_onerror(caller: Callable, path_str: PathName, excinfo: Any) -> None:
    """The exception handler used by rmtree_hard to try to remove read-only attributes.

    Args:
        caller: The calling function.
        path_str: The path to change.
        excinfo: The run stack to use if the caller is not remove or unlink.

    Returns:
        Nothing.

    Raises:
        The exception in excinfo if the caller is not remove or unlink.
    """
    path_str = Path(path_str)
    if caller not in (remove, unlink):
        raise excinfo[0](excinfo[1])
    path_str.chmod(S_IRWXU)
    path_str.unlink()


def _construct_remote_driver(remote: Optional[bool | str] = False, remote_is_windows: Optional[bool] = None, copy_for_remote: bool = False,  # pylint: disable=too-many-locals,too-many-branches
                             remote_auth: Optional[ServerAuthType] = None, remote_powershell: bool = False) -> List[str]:
    """Construct the remote command execution prefix for syscmd.

    Args:
        remote=False
        remote_is_windows=None
        copy_for_remote=False
        remote_auth=False
        remote_powershell=False

    Returns:
        The list which specifies the prefix for executing a remote command.

    Raises:
        CMDError.INVALID_OPERATION:
            If PowerShell remoting is requested while trying to pass remote credentials,
            If remote execution requires the executable to be copied before execution and the local or remote system is Linux.
            If remote execution requires the executable to be copied before execution and PowerShell remoting is requested.
    """
    remote_is_windows = WIN32 if (remote_is_windows is None) else remote_is_windows
    if WIN32:
        if remote_is_windows:
            if remote_powershell:
                if remote_auth or copy_for_remote:
                    raise CMDError(CMDError.INVALID_OPERATION, func='remote_auth' if remote_auth else 'copy_for_remote', context='PowerShell')
                remote_driver = ['powershell', '-NoLogo', '-NonInteractive', '-Command', 'Invoke-Command', '-ComputerName', str(remote), '-ScriptBlock']
            else:
                remote_driver = ['psexec', rf'\\{remote}', '-accepteula', '-nobanner', '-h']
                if isinstance(remote_auth, tuple):
                    remote_driver += ['-u', remote_auth[0], '-p', remote_auth[1]]
                if copy_for_remote:
                    remote_driver.append('-c')
                    copy_for_remote = False
        else:
            if copy_for_remote:
                raise CMDError(CMDError.INVALID_OPERATION, func='copy_for_remote', context='Linux')
            remote_driver = ['plink', '-batch', '-v']
            if isinstance(remote_auth, tuple):
                remote_driver += ['-l', remote_auth[0], '-pw', remote_auth[1]]
            elif remote_auth:
                remote_driver += ['-i', remote_auth]  # type: ignore[list-item]
            remote_driver += [str(remote)]
    else:
        if copy_for_remote:
            raise CMDError(CMDError.INVALID_OPERATION, func='copy_for_remote', context='Linux')
        remote_driver = ['ssh', '-t', '-t']
        if isinstance(remote_auth, tuple):
            remote_driver += ['-J', f'{remote_auth[0]}@{remote_auth[1]}']
        elif remote_auth:
            remote_driver += ['-i', remote_auth]  # type: ignore[list-item]
        remote_driver += [str(remote)]
    return remote_driver


def syscmd(command: PathName, /, *cmd_args, input_lines: Optional[Iterable] = None,  # pylint: disable=too-many-locals,too-many-branches,too-many-statements,too-many-arguments
           show_stdout: bool = False, ignore_stderr: bool = False, append_stderr: bool = False, fail_on_error: bool = True,
           show_cmd: bool = False, use_shell: bool = False, flatten_output: bool = False,
           remote: Optional[bool | str] = False, remote_is_windows: Optional[bool] = None, copy_for_remote: bool = False,
           remote_auth: Optional[ServerAuthType] = None, remote_powershell: bool = False) -> CommandResult:
    """Wrapper to provide a better interface to subprocess.Popen().

    Args:
        command
        *cmd_args
        input_lines=None
        show_stdout=False
        ignore_stderr=False
        append_stderr=False
        fail_on_error=True
        show_cmd=False
        use_shell=False
        flatten_output=False
        remote=False
        remote_is_windows=None
        copy_for_remote=False
        remote_auth=False
        remote_powershell=False

    Returns:
        The string (or string-list) output of the command.

    Raises:
        CMDError.INVALID_OPERATION: If any remote options are provided but remote is False.
        CMDError.CMD_NOT_FOUND: If the requested command is not found.
        CMDError.CMD_ERROR: If fail_on_error is True, and the return code is non-zero, or there is output on stderr if ignore_stderr is True.
    """
    cmd_spec = [str(command)] + [str(c) for c in cmd_args]
    remote_driver: List[str] = []
    if (not remote) and (remote_auth or copy_for_remote or remote_powershell or remote_is_windows):
        raise CMDError(CMDError.INVALID_OPERATION, func='remote options', context='local servers')
    if remote and WIN32 and remote_is_windows and not remote_powershell:
        ignore_stderr = True  # psexec puts status info on stderr
    if remote:
        remote_driver = _construct_remote_driver(remote, remote_is_windows, copy_for_remote, remote_auth, remote_powershell)
        remote_cmd = cmd_spec
        if use_shell and remote_is_windows:
            remote_cmd = ['cmd', '/c'] + cmd_spec
        use_shell = False
        if remote_powershell:
            remote_cmd = ['{'] + remote_cmd + ['}']
        cmd_spec = remote_driver + remote_cmd

    cmd_str = ' '.join([f'"{c}"' for c in cmd_spec])
    if show_cmd or is_debug('SYSCMD'):
        print('Executing system command:', cmd_str)
    cmd_to_pass = cmd_str if use_shell else cmd_spec

    with Popen(cmd_to_pass, shell=use_shell, universal_newlines=True, stdout=PIPE, stderr=PIPE,
               bufsize=0 if show_stdout else -1,
               stdin=PIPE if input_lines else None) as proc:
        if input_lines:
            if is_debug('SYSCMD'):
                print('Sending input lines:', input_lines)
            cast(IO, proc.stdin).writelines(input_lines)
            cast(IO, proc.stdin).close()

        if is_debug('SYSCMD'):
            print('Getting output lines')
        outlines = []
        for line in iter(cast(IO, proc.stdout).readline, ''):
            if is_debug('SYSCMD'):
                print('Received output line:', line)
            outlines.append(line)
            if show_stdout:
                sys.stdout.write(line)

        if is_debug('SYSCMD'):
            print('Getting error lines')
        err_lines = cast(IO, proc.stderr).readlines()
        if is_debug('SYSCMD'):
            print('Received error lines:', err_lines)

        if is_debug('SYSCMD'):
            print('Getting return code')
        try:
            returncode = proc.wait()
        except OSError as err:
            if err.errno == ECHILD:
                returncode = proc.returncode
            else:
                raise

    if is_debug('SYSCMD'):
        print('Received return code:', returncode)

    if remote and WIN32 and (not remote_is_windows) and (not returncode) and (err_lines[-1] == 'Disconnected: All channels closed\n'):
        err_lines = []

    if is_debug('SYSCMD'):
        print('Checking for' if fail_on_error else 'Ignoring', 'errors')
    if fail_on_error and (returncode or (not ignore_stderr and err_lines)):
        bad_cmd_str = 'not recognized as an internal or external command' if WIN32 else 'command not found'
        err_type = CMDError.CMD_NOT_FOUND if (bad_cmd_str in ''.join(err_lines)) else CMDError.CMD_ERROR
        raise CMDError(err_type, cmd=cmd_str, returncode=returncode, err_lines=err_lines, outlines=outlines)

    if err_lines and append_stderr:
        outlines += err_lines
    return flatten_string_list(outlines) if flatten_output else outlines


# Implement standard directory stack on chdir
_DIRECTORY_STACK = []


def pushd(dirname: PathName, /) -> PathName:
    """Implements the push function for a directory stack.

    Args:
        dirname: The directory to which to change.

    Returns:
        The value of the directory pushed to the stack.
    """
    global _DIRECTORY_STACK  # noqa:F824  # pylint: disable=global-variable-not-assigned
    cwd = Path.cwd()
    chdir(dirname)
    _DIRECTORY_STACK.append(cwd)
    return cwd


def popd() -> int | PathName:
    """Implements the pop function for a directory stack.

    Args:
        None.

    Returns:
        The value of the directory removed from the stack.

    Raises:
        IndexError: If the stack is empty.
    """
    global _DIRECTORY_STACK    # noqa:F824  # pylint: disable=global-variable-not-assigned
    try:
        dirname = _DIRECTORY_STACK.pop()
    except IndexError:
        return 0
    chdir(dirname)
    return dirname

# cSpell:ignore chgrp geteuid getpwnam IRGRP IROTH IRWXG IXOTH lockf NBLCK nobanner psexec pylance syscmd unlck ungrouped accepteula onexc



"""This module provides a simplified interface to the standard argparse module."""

# Import standard modules
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, HelpFormatter, Namespace, REMAINDER, _MutuallyExclusiveGroup
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, Iterable, Optional, Sequence, Type

# Import internal modules
from .version import AppVersion, VersionStyle
from .lang import str_to_pythonval


class Argument:  # pylint: disable=too-few-public-methods
    """This is a simple container class to encapsulate an argument definition passed to ArgumentParser.add_argument()."""

    def __init__(self, *names, **options) -> None:
        """
        Args:
            *names: A list of the argument names.
            **options (optional): A dictionary of the argument options.

        Attributes:
            names: The value of the names argument.
            options: The value of the options argument.
        """
        self.names: Sequence[str] = names
        self.options: Dict = options


@dataclass(frozen=True)
class SubParser:
    """This is a simple container class to encapsulate a subparser definition.

        Attributes:
            subcommand: The subcommand name for the subparser.
            command_runner: The function which runs the commands for the subparser.
            arguments (optional, default=None): A list of arguments for the subparser.
    """
    subcommand: str
    command_runner: Callable
    arguments: Sequence[Argument] = field(default_factory=tuple)


class Commander:
    """This class provides a simplified interface to the argparse.ArgumentParser class."""

    def __init__(self, description: str, arguments: Sequence[Argument] = tuple(),  # pylint: disable=too-many-locals,too-many-arguments,too-many-positional-arguments
                 subparsers: Iterable[SubParser] = tuple(), subparser_common_args: Sequence[Argument] = tuple(),
                 default: Optional[Callable] = None, parents: Sequence[ArgumentParser] = tuple(),
                 parse_extra: bool = False, extra_var_sep: str = ':', convert_extra: bool = True,
                 version: Optional[AppVersion] = None, version_style: VersionStyle = VersionStyle.one_line,
                 formatter_class: Type[HelpFormatter] = ArgumentDefaultsHelpFormatter):
        """
        Args:
            description: The subcommand name for the subparser.
            arguments (optional, default=None): A list of arguments for the command driver.
            subparsers (optional, default=None): A list of subparsers for the command driver.
            subparser_common_args (optional, default=None): A list of common arguments for all subparsers.
            default (optional, default=None): The default subparser.
            parents (optional, default=None): A list of parents to pass to the parser.
            parse_extra (optional, default=False): If True then all un-parsed arguments will be interpreted
                using the value of extra_var_sep as extra arguments of the form: argument:value.
            extra_var_sep (optional, default=:): The character used to separate the extra vars from the values.
            convert_extra (optional, default=True): If True then used str_to_pythonval on the values in the parse_extra arguments.
            version (optional, default=None): If not None then add a version argument to the parser and use AppVersion.get_info.
            version_style (optional, default=one_line): The version.VERSION_STYLE to use for the version argument output.
            formatter_class (optional, default=ArgumentDefaultsHelpFormatter): The formatter class to pass to the parser.

        Attributes:
            parser: The command parser instance.
            subparser_common_parser: The parser for subparser common arguments.
            _convert_extra: The value of the convert_extra argument.
            _default: The value of the default argument.
            _extra_var_sep: The value of the extra_var_sep argument.
            _parse_extra: The value of the parse_extra argument.
            _pass_on: The list of arguments to pass to the subparsers when parsing.
            _subparsers: The value of the subparsers argument.
        """
        self._convert_extra = convert_extra
        self._default = default
        self._extra_var_sep = extra_var_sep
        self._parse_extra = parse_extra
        self._pass_on: Sequence[str] = tuple()
        self._subparsers = subparsers
        self.subparser_common_parser: Optional[ArgumentParser] = None

        self.parser = ArgumentParser(description=description, formatter_class=formatter_class, parents=parents)  # type: ignore[arg-type]
        _add_arguments_to_parser(self.parser, arguments)
        if version:
            self.parser.add_argument('-v', '--version', action='version', version=version.get_info(version_style))

        if subparsers:
            subparser_objects = self.parser.add_subparsers(dest='command')
            for subparser_def in subparsers:
                subparser: ArgumentParser = subparser_objects.add_parser(subparser_def.subcommand)
                _add_arguments_to_parser(subparser, subparser_def.arguments)
                subparser.set_defaults(command=subparser_def.command_runner)

        if subparser_common_args:
            self.subparser_common_parser = ArgumentParser(add_help=False)
            _add_arguments_to_parser(self.subparser_common_parser, subparser_common_args)

        if self._parse_extra:
            self.parser.add_argument('extra_parser_args', nargs=REMAINDER, metavar='[[var1:val1] ...]')

    def execute(self, argv: Optional[Sequence[str]] = None, use_args: Optional[Namespace] = None) -> Any:
        """Parse the command line and call the command_runner.

        Args:
            argv (optional, default=None): The arguments to pass to the parser if use_args is None, otherwise sys.argv will be used.
            use_args (optional, default=None): The arguments to pass to the parser, otherwise argv will be used.

        Returns:
            The result of the called command_runner.
        """
        args: Namespace = use_args if use_args else self.parse_args(argv)
        caller: Callable = self._default if (self._subparsers and not args.command and self._default) else args.command
        caller_args = (self.subparser_common_parser, self._pass_on) if self.subparser_common_parser else (args,)
        return caller(*caller_args)

    def parse_args(self, argv: Optional[Sequence[str]] = None, err_msg: str = 'No command specified', raise_on_error: Optional[BaseException] = None) -> Namespace:
        """Parse the command line.

        Args:
            argv (optional, default=None): The arguments to pass to the parser, otherwise sys.argv will be used.
            err_msg (optional, default='No command specified'): The error message to use if a bad subcommand is requested.
            raise_on_error (optional, default=None): If not None, use the value as the error to be raised on a failure.

        Returns:
            The parsed argument Namespace.

        Raises:
            The value of raise_on_error if not False.
        """
        (args, self._pass_on) = self.parser.parse_known_args(argv) if self.subparser_common_parser else (self.parser.parse_args(argv), tuple())
        if self._subparsers and not args.command and not self._default:
            if raise_on_error:
                raise raise_on_error
            self.parser.error(err_msg)
        if self._parse_extra:
            for arg in args.extra_parser_args:
                if self._extra_var_sep in arg:
                    (var, val) = arg.split(self._extra_var_sep, 1)
                    setattr(args, var, str_to_pythonval(val) if self._convert_extra else val)
        return args


def _add_arguments_to_parser(parser: ArgumentParser | _MutuallyExclusiveGroup, arglist: Iterable[dict | Argument], /) -> None:
    """Add arguments to the specified parser.

    Args:
        parser: The parser to which to add the arguments.
        arglist: The arguments to add to the parser.

    Returns:
        Nothing.
    """
    for arg in arglist:
        if isinstance(arg, dict):
            group_parser = parser.add_mutually_exclusive_group(**arg['options'])
            _add_arguments_to_parser(group_parser, arg['args'])
        else:
            parser.add_argument(*arg.names, **arg.options)

# cSpell:ignore pythonval



"""This module provides a generic interface for local and remote server paths."""

# Import standard modules
from os import walk
from pathlib import Path, PurePosixPath, PureWindowsPath, WindowsPath
from string import Template
from shutil import copy
from typing import cast, Iterator, List, Optional, Tuple, TYPE_CHECKING

# Import internal modules
from .platarch import OsType
from .sysutil import rmpath, syscmd, CMDError
from .lang import is_debug, BatCaveError, BatCaveException, CommandResult, PathName, WIN32
if TYPE_CHECKING:
    from .servermgr import Server


class ServerPathError(BatCaveException):
    """ServerPath Exceptions.

    Attributes:
        INVALID_OPERATION: The specified server type does not support the requested operation.
        REMOTE_COPY_SPACE_ERROR: The remote server did not have enough disk space for the copy.
    """
    INVALID_OPERATION = BatCaveError(1, 'This function is only supported for remote Windows servers from Windows servers')
    REMOTE_COPY_SPACE_ERROR = BatCaveError(2, Template('Error during robocopy, possible lack of disk space on $dest'))


class ServerPath:
    """Class to create a universal abstract interface for an OS directory such that remote and local paths can be managed with the same code.

    Attributes:
        DEFAULT_REMOTE_COPY_COMMAND: The default command to perform a remote copy based on the OS of the source.
        DEFAULT_REMOTE_COPY_ARGS: The default arguments used to perform a remote copy based on the OS of the source.
    """
    DEFAULT_REMOTE_COPY_ARGS = {OsType.windows: ['/MIR', '/MT', '/R:0', '/NFL', '/NDL', '/NP', '/NJH', '/NJS'],
                                OsType.linux: ['-r', '-batch']}
    DEFAULT_REMOTE_COPY_COMMAND = {OsType.windows: 'robocopy', OsType.linux: 'pscp' if WIN32 else 'scp'}

    def __init__(self, server: 'Server', the_path: PathName, /):
        """
        Args:
            server: The server for which the file path is a reference.
            the_path: The file path on the server.

        Attributes:
            local: The .
            _raw_path: The value of the the_path argument.
            _server: The value of the server argument.
            win_to_win: .
        """
        self._server = server
        self._raw_path = the_path

    def __str__(self):
        return str(self.remote)

    def __truediv__(self, other: str):
        return ServerPath(self.server, self.local / other)

    is_win = property(lambda s: s._server.os_type == OsType.windows, doc='A read-only property which returns True if the path is on a Windows server.')
    local = property(lambda s: s.path_type(s._raw_path), doc='A read-only property which returns the value of the the_path argument when referenced locally on the server.')
    parent = property(lambda s: ServerPath(s.server, s.local.parent), doc='A read-only property which returns the parent of the path.')
    path_type = property(lambda s: PureWindowsPath if s.is_win else PurePosixPath, doc='A read-only property which returns the path type.')
    server = property(lambda s: s._server, doc='A read-only property which returns the owning server of the path.')
    win_to_win = property(lambda s: WIN32 and s.is_win, doc='A read-only property which returns True if the both the local and remote systems are Windows servers.')

    @property
    def remote(self) -> PathName:
        """A read-only property which returns the name of this remote server which hosts the path."""
        if self.server.is_local:
            return self.local
        if self.win_to_win:
            return WindowsPath(f'//{self.server.fqdn}/{self.local}'.replace(':', '$'))
        return f'{self.server.fqdn}:{self.local}'

    def copy(self, sp_dest: 'ServerPath', /, remote_cp_command: Optional[str] = None, remote_cp_args: Optional[List[str]] = None) -> CommandResult:
        """Implementation of shutil.copy() adding remote server support.

        Args:
            sp_dest: The destination of the copy.
            remote_cp_command (optional, default=None): If not None, the command to use if the path is remote other than the default.
            remote_cp_args (optional, default=None): If not None, the arguments to use if the path is remote other than the default.

        Returns:
            The result of the copy.

        Raises:
            ServerPathError.REMOTE_COPY_SPACE_ERROR: If the remote destination is out of space.
        """
        if self.win_to_win and WindowsPath(self.remote).is_dir():
            remote_cp_command = self.DEFAULT_REMOTE_COPY_COMMAND[OsType.windows] if (remote_cp_command is None) else remote_cp_command
            remote_cp_args = self.DEFAULT_REMOTE_COPY_ARGS[OsType.windows] if (remote_cp_args is None) else remote_cp_args

            dest: PathName
            if sp_dest.server.is_local and not self.server.is_local:
                use_server = sp_dest.server
                source = self.remote
                dest = sp_dest.local
            else:
                use_server = self.server
                source = self.local
                dest = sp_dest.local if (self.server == sp_dest.server) else sp_dest.remote

            try:
                return use_server.run_command(remote_cp_command, source, dest, *remote_cp_args, use_shell=True)
            except CMDError as err:
                if remote_cp_command != self.DEFAULT_REMOTE_COPY_COMMAND[OsType.windows]:
                    raise
                if 'returncode' in err.vars:
                    if err.vars['returncode'] in (1, 2, 3):
                        return ''
                    if err.vars['returncode'] in (8, 9):
                        raise ServerPathError(ServerPathError.REMOTE_COPY_SPACE_ERROR, dest=sp_dest) from err
                raise
        elif sp_dest.server.os_type == OsType.linux:
            remote_cp_command = self.DEFAULT_REMOTE_COPY_COMMAND[OsType.linux] if (remote_cp_command is None) else remote_cp_command
            remote_cp_args = self.DEFAULT_REMOTE_COPY_ARGS[OsType.linux] if (remote_cp_args is None) else remote_cp_args
            return syscmd(remote_cp_command, *remote_cp_args, self.local, sp_dest.remote)

        # if you get here it is a file copy from Windows to Windows, just use shutil
        return copy(self.remote, sp_dest.remote)

    def exists(self) -> bool:
        """Implementation of pathlib.Path.exists() adding remote server support.

        Returns:
            True if this path exists, False otherwise.
        """
        if self.server.is_local:
            if is_debug('SERVERPATH'):
                print(f'Testing local path: {self.local}')
            return Path(self.local).exists()
        if self.win_to_win:
            if is_debug('SERVERPATH'):
                print(f'Testing remote path: {self.remote}')
            return cast(Path, self.remote).exists()
        try:
            if is_debug('SERVERPATH'):
                print(f'Testing {self.server} remote path: {self.remote}')
            self.server.run_command('dir' if self.is_win else 'ls', self.local)
            return True
        except CMDError as err:
            if err.vars['err_lines'][0].startswith('ls: cannot access'):
                return False
            raise

    def iterdir(self) -> CommandResult | List[Path]:
        """Implementation of pathlib.Path.iterdir() adding remote server support.

        Returns:
            The contents of this directory path.
        """
        if self.server.is_local:
            return [i for i in Path(self.local).iterdir()]  # pylint: disable=unnecessary-comprehension
        if self.win_to_win:
            return [i for i in cast(Path, self.remote).iterdir()]  # pylint: disable=unnecessary-comprehension
        return self.server.run_command('dir' if self.is_win else 'ls', self.local)

    def mkdir(self, mode: int = 0o777, /, *, parents: bool = False, exist_ok: bool = False) -> None:
        """Implementation of pathlib.Path.mkdir() adding remote server support.

        Args:
            mode (optional, default=0o777): The access mode of the created directory.
            parents (optional, default=False): If True, also create intermediate directories.
            exist_ok (optional, default=False): If True, do not throw an exception if the path already exists.

        Returns:
            Nothing.
        """
        if self.server.is_local:
            Path(self.local).mkdir(mode, parents, exist_ok)
        if self.win_to_win:
            cast(Path, self.remote).mkdir(mode, parents, exist_ok)
        cmd: List[PathName] = ['mkdir']
        if parents and not self.is_win:
            cmd.append('-p')
        cmd.append(self.local)
        self.server.run_command(*cmd)

    def rename(self, new: str, /) -> CommandResult | PathName:
        """Implementation of pathlib.Path.rename() adding remote server support.

        Args:
            new: The new path name.

        Returns:
            The result of the rename.
        """
        if self.server.is_local:
            return Path(self.local).rename(new)
        if self.win_to_win:
            return cast(Path, self.remote).rename(new)
        return self.server.run_command('ren' if self.is_win else 'mv', self.local, new)

    def rmdir(self, remote_rm_command: Optional[List[PathName]] = None, *, recursive: bool = False) -> None:
        """Implementation of pathlib.Path.rmdir() adding remote server support.

        Args:
            remote_rm_command (optional, default=None): If not None, the command to use if the path is remote other than the default.
            recursive (optional, default=False): If True, also remove all subdirectories.

        Returns:
            Nothing.
        """
        if remote_rm_command is None:
            remote_rm_command = ['RD', '/Q'] if self.is_win else ['rm', '-f']

        if not recursive:
            if self.server.is_local:
                Path(self.local).rmdir()
            if self.win_to_win:
                cast(Path, self.remote).rmdir()
            remote_rm_command.append(self.local)
            self.server.run_command(*remote_rm_command, use_shell=True)

        if self.server.is_local:
            rmpath(self.local)
        if self.win_to_win:
            rmpath(self.remote)
        remote_rm_command += ['/S' if self.is_win else '-r', self.local]
        self.server.run_command(*remote_rm_command, use_shell=True)

    def walk(self) -> Iterator[Tuple[str, List[str], List[str]]]:
        """Implementation of os.walk() adding remote server support.

        Returns:
            The result of the walk.

        raises:
            ServerPathError.INVALID_OPERATION: If the command is not supported on the target path.
        """
        if self.server.is_local:
            return walk(self.local)
        if self.win_to_win:
            return walk(self.remote)
        raise ServerPathError(ServerPathError.INVALID_OPERATION)

# cSpell:ignore syscmd platarch servermgr serverpath
